function [] = usrpBuildBinaries(additionalMexOptions)
% @version:		$Id: usrpBuildBinaries.m 220 2009-05-15 07:39:45Z nagel $
% @author:		Enno Klasing <klasing@int.uni-karlsruhe.de>
% @copyright:	Institut fr Nachrichtentechnik, http://www.int.uni-karlsruhe.de
% @license:		GNU General Public License, Version 3
%
% additionalMexOptions can be a cell array with option strings which will be passed to mex.
% To built optimizd binaries, use the '-O' option:
% >> usrpBuildBinaries({'-O'})
% To built debugable binaries, use the '-g' option:
% >> usrpBuildBinaries({'-g'})

% Check input arguments
error(nargchk(0, 1, nargin, 'struct'));
if nargin < 1
	additionalMexOptions = {};
else
	if ~iscell(additionalMexOptions) || size(additionalMexOptions, 2) > 1
		error('additionalMexOptions must be a column cell array with one column');
	end
end

% Root path for all source files
binPath = fileparts(mfilename('fullpath'));
projectPath = fullfile(fileparts(binPath), 'src');

% Output directory
additionalMexOptions(end+1, 1) = {['-outdir ''' binPath '''']};

% Clean up old debug files, if present
delete(fullfile(binPath, '*.pdb'));

%% Compile library "libusrp"
mexOptions = [
	additionalMexOptions
	'-output libusrp'
	% This is a hack to make mex() use Microsofts library linker instead of "link.exe",
	% which is used normally for building DLLs
	'LINKER=lib.exe'
	'LINKFLAGS=/NODEFAULTLIB'
];
sourcePath = fullfile(projectPath, 'usrp', 'libusrp');
sourceFiles = {
	'db_base.cpp'
	'db_basic.cpp'
	'db_boards.cpp'
	'db_dbs_rx.cpp'
	'db_dtt754.cpp'
	'db_dtt768.cpp'
	'db_flexrf.cpp'
	'db_flexrf_mimo.cpp'
	'db_tv_rx.cpp'
    'db_tv_rx_mimo.cpp'
	'db_util.cpp'
    'db_wbxng.cpp'
    'db_wbxng_adf4350.cpp'
    'db_wbxng_adf4350_regs.cpp'
	'db_xcvr2450.cpp'
	'fusb.cpp'
	'fusb_sysconfig_win32.cpp'
	'fusb_win32.cpp'
	'md5.c'
	'usrp_basic.cpp'
	'usrp_config.cpp'
	'usrp_dbid.cpp'
	'usrp_prims.cpp'
    'usrp_prims_libusb0.cpp'
	'usrp_standard.cpp'
	'libusb_dyn.c'
};
includePaths = {
	sourcePath
	fullfile(sourcePath, 'include')
	fullfile(sourcePath, 'include', 'usrp')
	fullfile(projectPath, 'includes')
};
usrpPreCompileEvent(binPath, 'libusrp');
usrpCallMex(true, mexOptions, sourcePath, sourceFiles, includePaths);
usrpPostCompileEvent(binPath, 'libusrp');

%% Compile library "simulink-usrp"
mexOptions = [
	additionalMexOptions
	'-output ''simulink-usrp'''
	% This is a hack to make mex() use Microsofts library linker instead of "link.exe",
	% which is used normally for building DLLs
	'LINKER=lib.exe'
	'LINKFLAGS=/NODEFAULTLIB'
];
sourcePath = fullfile(projectPath, 'simulink-usrp', 'simulink-usrp');
sourceFiles = {
	'MatlabArrayHelper.cpp'
	'MatlabFunctionHelper.cpp'
	'MatlabUsrpHelper.cpp'
	'SimulinkUsrpHelper.cpp'
	'UsrpHelper.cpp'
	'UsrpSimulinkSink.cpp'
	'UsrpSimulinkSource.cpp'
};
includePaths = {
	fullfile(projectPath, 'usrp', 'libusrp')
	fullfile(projectPath, 'usrp', 'libusrp', 'include')
	fullfile(projectPath, 'usrp', 'libusrp', 'include', 'usrp')
	fullfile(projectPath, 'includes')
    fullfile(matlabroot,'simulink','include')
};
usrpPreCompileEvent(binPath, 'simulink-usrp');
usrpCallMex(true, mexOptions, sourcePath, sourceFiles, includePaths);
usrpPostCompileEvent(binPath, 'simulink-usrp');

%% Compile S-Functions usrp_helper, usrp_sink, usrp_source
% Same settings for all of these:
includePaths = [
	fullfile(projectPath, 'simulink-usrp', 'simulink-usrp')
	includePaths
];
libFiles = {
	'libusrp'
	'simulink-usrp'
};
libPaths = {
	binPath
};

% usrp-helper
sourcePath = fullfile(projectPath, 'simulink-usrp', 'simulink-usrp-helper');
sourceFiles = {
	'usrp_helper.cpp'
	'UsrpHelperMexFcn.cpp'
	'UsrpHelperSFcnCallbacks.cpp'
};
usrpCallMex(true, additionalMexOptions, sourcePath, sourceFiles, includePaths, libFiles, libPaths);

% usrp-sink
sourcePath = fullfile(projectPath, 'simulink-usrp', 'simulink-usrp-sink');
sourceFiles = {'usrp_sink.cpp'};
usrpCallMex(true, additionalMexOptions, sourcePath, sourceFiles, includePaths, libFiles, libPaths);

% usrp-source
sourcePath = fullfile(projectPath, 'simulink-usrp', 'simulink-usrp-source');
sourceFiles = {'usrp_source.cpp'};
usrpCallMex(true, additionalMexOptions, sourcePath, sourceFiles, includePaths, libFiles, libPaths);

%% Compile stdout2matlab
sourcePath = fullfile(projectPath, 'simulink-usrp', 'stdout2matlab');
sourceFiles = {
	'stdout2matlab.cpp'
	'log2namedPipe.cpp'
};
usrpCallMex(true, additionalMexOptions, sourcePath, sourceFiles);

%% Configure the helper block


open usrp_GPP;
set_param('usrp_GPP','Lock','off')
set_param('usrp_GPP/usrp_helper','firmware','');
set_param('usrp_GPP/usrp_helper','bitstream','');
set_param('usrp_GPP','Lock','on');
save_system usrp_GPP;
close_system usrp_GPP;

%% Clean up
delete(fullfile(binPath, '*.ilk'));
% ^^ Incremental linker files. Not used anyway, so throw them away...
% Btw, MATLAB 2009a does this by itself already.

end


%% Create a mex-Command and run it
function [] = ...
	usrpCallMex(realRun, mexSwitches, sourcePath, sourceFiles, includePaths, libFiles, libPaths)

% Check input arguments
error(nargchk(4, 7, nargin, 'struct'));
if nargin < 5
	includePaths = [];
end
if nargin < 6
	libFiles = [];
end
if nargin < 7
	libPaths = [];
end

% Built a command line for calling mex
mexCommand = 'mex';

% Mex Switches
for mexSwitch = mexSwitches'
	mexCommand = [mexCommand, ' ', char(mexSwitch)]; %#ok<AGROW>
end

% Source Files
for sourceFile = sourceFiles'
	mexCommand = strcat(mexCommand, ' ''', fullfile(sourcePath, char(sourceFile)), '''');
end

% Include Paths
for includePath = includePaths'
	mexCommand = strcat(mexCommand, ' -I''', char(includePath), '''');
end

% Library Paths
for libPath = libPaths'
	mexCommand = strcat(mexCommand, ' -L''', char(libPath), '''');
end

% Library Files
for libFile = libFiles'
	mexCommand = strcat(mexCommand, ' -l''', char(libFile), '''');
end

% Call Mex
if realRun
	disp(['Running: ' mexCommand]);
	eval(mexCommand);
else
	disp(['Dry run: ' mexCommand]);
end

end

%%
% Creates files which mex might want to delete after creating a mex-File, but which will not be
% created when using lib.exe instead of link.exe
function [] = usrpPreCompileEvent(binPath, libName)

% Check input arguments
error(nargchk(2, 2, nargin, 'struct'));

% Create files
temp = 0; %#ok<NASGU>
save(fullfile(binPath, [libName '.' mexext() '.manifest']), 'temp');
save(fullfile(binPath, [libName '.map']), 'temp'); % R2008a
save(fullfile(binPath, [libName '.' mexext() '.map']), 'temp'); % R2009a

end

%%
% Renames files after using mex for creating libraries instead of mex-Files, and deletes any files
% which might be left over by mex
function [] = usrpPostCompileEvent(binPath, libName)

% Check input arguments
error(nargchk(2, 2, nargin, 'struct'));

% Rename files
for file = dir(fullfile(binPath, [libName '.' mexext() '*']))'
	[path, name, ext] = fileparts(file.name);
	if strcmp(ext, ['.' mexext()])
		ext = '.lib';
	end
	movefile(fullfile(binPath, file.name), fullfile(binPath, [libName ext]));
end

% Delete files
for file = dir(fullfile(binPath, [libName '.manifest']))'
	delete(fullfile(binPath, file.name));
end
for file = dir(fullfile(binPath, [libName '.map']))'
	delete(fullfile(binPath, file.name));
end

end

