/**
 * usrp_helper MexFunction
 *
 * A MexFunction can be contained in a mex-File together with an S-Function by
 * calling it ProcessMexSfunctionCmdLineCall() and defining the makro
 * PROCESS_MEX_SFUNCTION_CMD_LINE_CALL. See also simulink.c at the end of
 * mexFunction().
 *
 * @version:	$Id: UsrpHelperMexFcn.cpp 220 2009-05-15 07:39:45Z nagel $
 * @author:		Enno Klasing <klasing@int.uni-karlsruhe.de>
 * @copyright:	Institut fr Nachrichtentechnik, http://www.int.uni-karlsruhe.de
 * @license:	GNU General Public License, Version 3
 */

#include "UsrpHelperMexFcn.h"

#include <string>
#include <sstream>
#include <MatlabArrayHelper.h>
#include <MatlabFunctionHelper.h>
#include <MatlabUsrpHelper.h>
#include <SimulinkUsrpHelper.h>
#include <UsrpHelper.h>

using namespace std;


/**
 * Helper function for the Simulink-USRP-Interface
 */
bool
ProcessMexSfunctionCmdLineCall(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	wstring functionName;

	/*
	 * Check if this is a Simulink call, or a Matlab command line call, by examing the parameters.
	 * prhs[0] contains the "time", prhs[3] contains the "mode flag".
	 */
	if (nrhs >= 4 && nlhs >= 1 &&
		mxIsDouble(prhs[0]) &&
		!mxIsComplex(prhs[0]) &&
		mxIsNumeric(prhs[3]) &&
		!mxIsComplex(prhs[3]) &&
		mxGetNumberOfElements(prhs[3]) == 1
	) {
		// Return false to let the mexFunction in simulink.c know that we did not handle this call
		return false;
	}

	/*
	 * Redirect stdout and stderr into Matlab. To do this only once, we use stdout2matlab()
	 * as one central place for doing this.
	 */
	mexCallMATLAB(0, NULL, 0, NULL, "stdout2matlab");

	// Set the environment variable USRP_PATH
	SimulinkUsrpHelper::setPathEnvironmentVar();

	// Get the name of the choosen function
	if (nrhs < 1) {
		mexErrMsgTxt("No sub-function specified.\n");
	}
	functionName = MatlabArrayHelper::getWStringFromArray(prhs[0], "sub_function");

	/*
	 * Check which function is being called and call the function.
	 * The function is responsible for checking the rest of the parameters.
	 */
	if (functionName == L"BoardInfo") {
		UsrpHelperMexFcn::BoardInfo(nlhs, plhs, nrhs, prhs);
	} else if (functionName == L"BoardAndSubdeviceInfo") {
		UsrpHelperMexFcn::BoardAndSubdeviceInfo(nlhs, plhs, nrhs, prhs);
	} else if (functionName == L"SamplingRateAndBandwidth") {
		UsrpHelperMexFcn::SamplingRateAndBandwidth(nlhs, plhs, nrhs, prhs);
	} else if (functionName == L"GetParam") {
		UsrpHelperMexFcn::GetParam(nlhs, plhs, nrhs, prhs);
	} else {
		static char errorMessage[256];
		_snprintf_s(errorMessage, _TRUNCATE, "No valid sub-function specified: %S.\n", functionName.c_str());
		mexErrMsgTxt(errorMessage);
	}

	return true;
}


/**
 * Writes a string with the serial number of the Nth USRP Board on this computer, and a string
 * with the name of all installed daughterboards, into the current Simulink Block.
 */
void
UsrpHelperMexFcn::BoardInfo(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	mxArray *gcb = NULL;
	mxArray *tempArray = NULL;

	// Check number of input and output parameters
	if (nrhs != 1 || nlhs != 0) {
		mexErrMsgTxt(__FUNCTION__ ": Wrong number of input or output parameters.\n");
	}

	// Do nothing if a simulation is running
	if (UsrpHelperMexFcn::SimulationIsRunning()) {
		return;
	}

	// Fetch path to the current block
	gcb = MatlabFunctionHelper::gcb();

	// Fetch the value of the parameter "usrp_number"
	tempArray = MatlabFunctionHelper::getParam(gcb, "usrp_number");
	MatlabFunctionHelper::str2Num(&tempArray);
	int boardNumber = MatlabArrayHelper::getIntFromArray(
		tempArray, "board_number", false, 0, 1023
	);
	mxDestroyArray(tempArray);

	// Fetch the board and daughterboard info
	string boardInfo = UsrpHelper::boardInfo(boardNumber);
	string rxDaughterboards, txDaughterboards;
	UsrpHelper::daughterboardInfo(boardNumber, rxDaughterboards, txDaughterboards);

	// Write the result into the "daughterboards" parameter
	vector <string> parameters;
	parameters.push_back("serial_number");
	parameters.push_back(boardInfo);
	parameters.push_back("daughterboards_rx");
	parameters.push_back(rxDaughterboards);
	parameters.push_back("daughterboards_tx");
	parameters.push_back(txDaughterboards);
	MatlabFunctionHelper::setParam(gcb, parameters);

	mxDestroyArray(gcb);
}


/**
 * Writes a string with the serial number of the Nth USRP Board on this computer, and a string
 * for each RX or TX subdevice, as well as the available gain ranges for each subdevice, into
 * the current Simulink block.
 */
void
UsrpHelperMexFcn::BoardAndSubdeviceInfo(
	int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]
) {
	mxArray *gcb = NULL;
	mxArray *tempArray = NULL;
	UsrpHelper::subdevice sideASubdev1, sideASubdev2, sideBSubdev1, sideBSubdev2;

	// Check number of input and output parameters
	if (nrhs != 1 || nlhs != 0) {
		mexErrMsgTxt(__FUNCTION__ ": Wrong number of input or output parameters.\n");
	}

	// Do nothing if a simulation is running
	if (UsrpHelperMexFcn::SimulationIsRunning()) {
		return;
	}

	// Fetch path to the current block
	gcb = MatlabFunctionHelper::gcb();

	// Get the type (RX/TX) of the current block
	tempArray = MatlabFunctionHelper::getParam(gcb, "FunctionName");
	string functionName = MatlabArrayHelper::getStringFromArray(tempArray, "Block Type");
	bool isTx = UsrpHelper::typeString2RxType(functionName);
	mxDestroyArray(tempArray);

	// Fetch the value of the parameter "usrp_number"
	tempArray = MatlabFunctionHelper::getParam(gcb, "usrp_number");
	MatlabFunctionHelper::str2Num(&tempArray);
	int boardNumber = MatlabArrayHelper::getIntFromArray(
		tempArray, "board_number", false, 0, 1023
	);
	mxDestroyArray(tempArray);

	// Fetch the board and subdevice info
	string serialNumber = UsrpHelper::boardInfo(boardNumber);
	UsrpHelper::subdeviceInfo(
		boardNumber, isTx, sideASubdev1, sideASubdev2, sideBSubdev1, sideBSubdev2
	);

	// Write the result into the corresponding parameters
	vector <string> parameters;
	parameters.push_back("serial_number");
	parameters.push_back(serialNumber);
	// Side A, Subdev 1
	parameters.push_back("sideA_subdev1_subdevice");
	parameters.push_back(sideASubdev1.name);
	parameters.push_back("sideA_subdev1_frequency_range");
	parameters.push_back(sideASubdev1.frequencyRange);
	parameters.push_back("sideA_subdev1_gain_range");
	parameters.push_back(sideASubdev1.gainRange);
	// Side A, Subdev 2
	if (!isTx) {
		parameters.push_back("sideA_subdev2_subdevice");
		parameters.push_back(sideASubdev2.name);
		parameters.push_back("sideA_subdev2_frequency_range");
		parameters.push_back(sideASubdev2.frequencyRange);
		parameters.push_back("sideA_subdev2_gain_range");
		parameters.push_back(sideASubdev2.gainRange);
	}
	// Side B, Subdev 1
	parameters.push_back("sideB_subdev1_subdevice");
	parameters.push_back(sideBSubdev1.name);
	parameters.push_back("sideB_subdev1_frequency_range");
	parameters.push_back(sideBSubdev1.frequencyRange);
	parameters.push_back("sideB_subdev1_gain_range");
	parameters.push_back(sideBSubdev1.gainRange);
	// Side B, Subdev 2
	if (!isTx) {
		parameters.push_back("sideB_subdev2_subdevice");
		parameters.push_back(sideBSubdev2.name);
		parameters.push_back("sideB_subdev2_frequency_range");
		parameters.push_back(sideBSubdev2.frequencyRange);
		parameters.push_back("sideB_subdev2_gain_range");
		parameters.push_back(sideBSubdev2.gainRange);
	}
	MatlabFunctionHelper::setParam(gcb, parameters);

	mxDestroyArray(gcb);
}


/**
 * Writes the sampling rate and used bandwidth for a given decimation factor and samples width
 * (8/16 Bit) into the current Simulink block.
 */
void
UsrpHelperMexFcn::SamplingRateAndBandwidth(
	int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]
) {
	mxArray *gcb = NULL;
	mxArray *tempArray = NULL;

	// Check number of input and output parameters
	if (nrhs != 1 || nlhs != 0) {
		mexErrMsgTxt(__FUNCTION__ ": Wrong number of input or output parameters.\n");
	}

	// Do nothing if a simulation is running
	if (UsrpHelperMexFcn::SimulationIsRunning()) {
		return;
	}

	// Fetch path to the current block
	gcb = MatlabFunctionHelper::gcb();

	// Get the type (RX/TX) of the current block
	tempArray = MatlabFunctionHelper::getParam(gcb, "FunctionName");
	string functionName = MatlabArrayHelper::getStringFromArray(tempArray, "Block Type");
	bool isTx = UsrpHelper::typeString2RxType(functionName);
	mxDestroyArray(tempArray);

	// Fetch the value of the parameter "decimation"
	tempArray = MatlabFunctionHelper::getParam(gcb, "decimation");
	MatlabFunctionHelper::str2Num(&tempArray);
	int decimation = MatlabArrayHelper::getIntFromArray(tempArray, "decimation");
	mxDestroyArray(tempArray);

	// Fetch the value of the parameter "sample_width"
	tempArray = MatlabFunctionHelper::getParam(gcb, "sample_width");
	string sampleWidthString = MatlabArrayHelper::getStringFromArray(tempArray, "sample_width");
	int sampleWidth = UsrpHelper::sampleWidthString2SampleWidth(sampleWidthString);
	mxDestroyArray(tempArray);

	// Find out how many active channels there are
	int channelCount = 0;
	char paramName[] = "sideX_subdevX_datatype";
	for (int i = 0; i < 4 /* TODO: constant */; i += isTx ? 2 : 1) {
		paramName[4] = 'A' + (i / 2);
		paramName[12] = '1' + i % 2;

		tempArray = MatlabFunctionHelper::getParam(gcb, paramName);
		wstring dataType = MatlabArrayHelper::getWStringFromArray(tempArray, paramName);
		mxDestroyArray(tempArray);

		if (dataType != L"Disabled") {
			channelCount++;
		}
	}

	// Fetch the sampling rate info
	string result = UsrpHelper::samplingRateAndBandwidthInfo(
		isTx, decimation, channelCount, sampleWidth
	);

	// Write the result into the "sampling_rate_and_bandwidth" parameter
	vector <string> parameters;
	parameters.push_back("sampling_rate_and_bandwidth");
	parameters.push_back(result);
	MatlabFunctionHelper::setParam(gcb, parameters);

	mxDestroyArray(gcb);
}


/**
 * Reads parameters from a usrp_helper block and returns them to Matlab.
 */
void
UsrpHelperMexFcn::GetParam(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	// Check number of input and output parameters
	if (nrhs != 2 || nlhs != 1) {
		mexErrMsgTxt(__FUNCTION__ ": Wrong number of input or output parameters.\n");
	}

	// Fetch path to the current model
	mxArray *bdroot = MatlabFunctionHelper::bdroot();

	// Fetch usrp_number entered into current block
	mxArray *gcb = MatlabFunctionHelper::gcb();
	mxArray *usrpNumber = MatlabFunctionHelper::getParam(gcb, "usrp_number");
	MatlabFunctionHelper::str2Num(&usrpNumber);
	mxDestroyArray(gcb);

	// Fetch param from associated helper block
	string paramName = MatlabArrayHelper::getStringFromArray(prhs[1], "parameter name");
	plhs[0] = MatlabUsrpHelper::fetchParamFromHelperBlock(
		bdroot, usrpNumber, paramName
	);
	mxDestroyArray(bdroot);
	mxDestroyArray(usrpNumber);

	// Return default value if no usrp_helper block is present
	if (paramName == "firmware" || paramName == "bitstream") {
		if (plhs[0] == NULL) {
			plhs[0] = mxCreateString("");
		}
	} else {
		if (plhs[0] == NULL && paramName == "sample_time_factor") {
			plhs[0] = mxCreateDoubleScalar(1);
		} else if (plhs[0] == NULL) {
			plhs[0] = mxCreateDoubleScalar(0);
		} else {
			MatlabFunctionHelper::str2Num(&plhs[0]);
		}
	}
}


/**
 * Checks whether the currently active model is running
 */
bool
UsrpHelperMexFcn::SimulationIsRunning()
{
	// Fetch path to the current model
	mxArray *bdroot = MatlabFunctionHelper::bdroot();

	// Get simulation status of the model
	mxArray *simulationStatusArray = MatlabFunctionHelper::getParam(bdroot, "SimulationStatus");
	string simulationStatus = MatlabArrayHelper::getStringFromArray(
		simulationStatusArray, "SimulationStatus"
	);
	mxDestroyArray(simulationStatusArray);
	mxDestroyArray(bdroot);

	// Check simulation status
	if (simulationStatus == "stopped" || simulationStatus == "initializing") {
		return false;
	} else {
		return true;
	}
}
