/**
 * Interface betweeen the USRP and Simulink, the USRP acts as a data source
 *
 * @version:	$Id: usrp_source.cpp 213 2009-05-13 09:03:24Z nagel $
 * @author:		Enno Klasing <klasing@int.uni-karlsruhe.de>
 * @copyright:	Institut fr Nachrichtentechnik, http://www.int.uni-karlsruhe.de
 * @license:	GNU General Public License, Version 3
 */

#define S_FUNCTION_LEVEL 2
#define S_FUNCTION_NAME usrp_source
#define USRP_TYPE UsrpSimulinkSource

/**
 * C++ Includes
 */
#include <UsrpSimulinkSource.h>
#include <MatlabArrayHelper.h>
#ifdef MATLAB_MEX_FILE
#include <MatlabFunctionHelper.h>
#endif
#include <SimulinkUsrpHelper.h>
#include <UsrpHelper.h>
#include <WhereAmI.h>
#include <fusb.h>
#include <string>
#include <simstruc.h>

using namespace std;

// ssSetErrorStatus() expects a static character array, that's why we define this error string
// in a file-wide variable.
static char errorMessage[512];	

/**
 * Enum containing the names of the parameters of this S-Function in correct ordering
 */
enum SFcnParamsIndex
{
	USRP_NUMBER = 0,
	DECIMATION,
	SAMPLE_WIDTH,
	VECTOR_LENGTH,
	SIDEA_SUBDEV1_DATATYPE,
	SIDEA_SUBDEV1_FREQUENCY,
	SIDEA_SUBDEV1_FREQUENCY_PORT,
	SIDEA_SUBDEV1_GAIN,
	SIDEA_SUBDEV1_ANTENNA,
	SIDEA_SUBDEV2_DATATYPE,
	SIDEA_SUBDEV2_FREQUENCY,
	SIDEA_SUBDEV2_FREQUENCY_PORT,
	SIDEA_SUBDEV2_GAIN,
	// Because parameter indices will also be calculated by offset from SIDEA_SUBDEV1_xxx, it is
	// important that every subdev has the same number of parameters
	SIDEA_SUBDEV2_DUMMYPARAM,
	SIDEB_SUBDEV1_DATATYPE,
	SIDEB_SUBDEV1_FREQUENCY,
	SIDEB_SUBDEV1_FREQUENCY_PORT,
	SIDEB_SUBDEV1_GAIN,
	SIDEB_SUBDEV1_ANTENNA,
	SIDEB_SUBDEV2_DATATYPE,
	SIDEB_SUBDEV2_FREQUENCY,
	SIDEB_SUBDEV2_FREQUENCY_PORT,
	SIDEB_SUBDEV2_GAIN,
	// Because parameter indexes will also be calculated by offset from SIDEA_SUBDEV1_xxx, it is
	// important that every subdev has the same number of parameters
	SIDEB_SUBDEV2_DUMMYPARAM,
	SAMPLE_TIME_FACTOR,
	USB_BLOCK_SIZE,
	USB_NBLOCKS,
	FIRMWARE,
	BITSTREAM,
	// SFCN_PARAMS_LENGTH must be the last in this enum to correctly set the number of expected parameters
	S_FCN_PARAMS_LENGTH
};
#define SUBDEV_PARAM_DISTANCE (SIDEA_SUBDEV2_DATATYPE - SIDEA_SUBDEV1_DATATYPE)
#define MAX_SUBDEV_COUNT 4
#define MAX_CHANNEL_COUNT MAX_SUBDEV_COUNT

/***
Variables used for dynamic frequency tuning
**/
int usrp_source_dynfreq_numInputPorts;
int usrp_source_dynfreq_subdevs[MAX_SUBDEV_COUNT];
double usrp_source_dynfreq_subdevs_tune_hold[MAX_SUBDEV_COUNT];
db_base_sptr usrp_source_dynfreq_subdevs_ptr[MAX_SUBDEV_COUNT];

/**
 * Enum containing the position of various pointers in the PWork vector
 */
enum PWorkIndex
{
	// USRP object
	USRP = 0,
	// Function pointers to functions which copy data from the USRP's local buffer to the
	// corresponding output port(s). The correct function to use for each port is determined once
	// in mdlStart to save complex if-statements in mdlOutputs.
	COPY_FUNCTION_POINTER0,
	COPY_FUNCTION_POINTER1,
	COPY_FUNCTION_POINTER2,
	COPY_FUNCTION_POINTER3,
	// P_WORK_LENGTH must be the last in this enum to correctly identify the length of the PWork
	// vector
	P_WORK_LENGTH
};
#define COPY_FUNCTION_POINTER_COUNT 4

/**************************************************************************************************
 * S-function methods
 **************************************************************************************************/

/**
 * mdlInitializeSizes
 */
static void
mdlInitializeSizes(SimStruct *S)
{
	// SS_SIMMODE_NORMAL = 0, ...SIZES_CALL_ONLY = 1, ...RTWGEN = 2, ...EXTERNAL = 3
	WHERE_AM_I(S, "Simulation mode: %d.", ssGetSimMode(S));

#ifdef MATLAB_MEX_FILE
	// Redirect stdout and stderr into Matlab. To do this only once, we use stdout2matlab()
	// as one central place for doing this.
	mexCallMATLAB(0, NULL, 0, NULL, "stdout2matlab");

	// Set the environment variable USRP_PATH
	SimulinkUsrpHelper::setPathEnvironmentVar();
#endif

	// Check number of S-Function Parameters
	ssSetNumSFcnParams(S, S_FCN_PARAMS_LENGTH);
	if (ssGetSFcnParamsCount(S) != S_FCN_PARAMS_LENGTH) {
		// The Simulink engine displays an error message by itself.
		return;
	}

	// Determine the number of ports
	vector <SimulinkUsrpHelper::subdevInfo> subdevs(MAX_SUBDEV_COUNT);
	int portCount = 0;
	int subdevCount = SimulinkUsrpHelper::countEnabledSubdevices(
		S, MAX_SUBDEV_COUNT, SIDEA_SUBDEV1_DATATYPE, SUBDEV_PARAM_DISTANCE,
		subdevs, &portCount
	);
	if (portCount == 0) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"All subdevices are disabled.\n"
		);
		ssSetErrorStatus(S, errorMessage);
		return;
	}
	
	string freq_port;
	usrp_source_dynfreq_numInputPorts = 0;

	// Initialize dynamic frequency tuning
	for (int i = 0; i < MAX_SUBDEV_COUNT; i++) {
		
		// Initialize dynamic frequency variables
		usrp_source_dynfreq_subdevs[i] = -1;
		usrp_source_dynfreq_subdevs_tune_hold[i] = -1;

		// if subdevice not active --> exit
		if (!subdevs[i].enabled) {
			continue;
		}

		// Check if frequency port is seleceted for subdev
		freq_port = MatlabArrayHelper::getStringFromArray(
			ssGetSFcnParam(S, SIDEA_SUBDEV1_FREQUENCY_PORT+i*SUBDEV_PARAM_DISTANCE), "sideX_subdevX_frequency_port"
		);

		// If selectet --> increase number of input ports		
		if(freq_port == "on"){
			
			usrp_source_dynfreq_subdevs[i] = usrp_source_dynfreq_numInputPorts;
			usrp_source_dynfreq_numInputPorts++;
		}
	}

	// Set number of input ports
	if (!ssSetNumInputPorts(S, usrp_source_dynfreq_numInputPorts)) {
		// The Simulink engine displays an error message by itself.
		return;
	}

	// Set number of output ports
	if (!ssSetNumOutputPorts(S, portCount)) {
		// The Simulink engine displays an error message by itself.
		return;
	}

	// Fetch input parameter 'vector_length'
	int vectorLength = MatlabArrayHelper::getIntFromArray(
		ssGetSFcnParam(S, VECTOR_LENGTH), "vector_length", false, 0
	);

	// Set the data type etc. on each port
	int portNum = 0;
	for (int i = 0; i < MAX_SUBDEV_COUNT; i++) {
		if (!subdevs[i].enabled) {
			continue;
		}

		// Get the number of the builtin datatype
		BuiltInDTypeId dataType;
		if (!SimulinkUsrpHelper::dataTypeString2BuiltinDataType(subdevs[i], dataType)) {
			_snprintf_s(errorMessage, _TRUNCATE, "Unknown 'data_type' parameter: '%S'.\n",
				subdevs[i].dataType.c_str()
			);
			ssSetErrorStatus(S, errorMessage);
			return;
		}

		// Set port parameters
		ssSetOutputPortComplexSignal(S, portNum, subdevs[i].complexOutput);
		ssSetOutputPortDataType(S, portNum, dataType);
		ssSetOutputPortFrameData(S, portNum, FRAME_YES);
		ssSetOutputPortMatrixDimensions(S, portNum, vectorLength, 1);
		portNum++;
		if (subdevs[i].complexOutput == COMPLEX_NO) {
			ssSetOutputPortComplexSignal(S, portNum, subdevs[i].complexOutput);
			ssSetOutputPortDataType(S, portNum, dataType);
			ssSetOutputPortFrameData(S, portNum, FRAME_YES);
			ssSetOutputPortMatrixDimensions(S, portNum, vectorLength, 1);
			portNum++;
		}
	}

	// Set the input port parameters if dynamic frequency is enabled
	for (int i=0;i<usrp_source_dynfreq_numInputPorts;i++){

        ssSetInputPortDirectFeedThrough(  S, i, 1);
        ssSetInputPortFrameData(          S, i, FRAME_NO);
		ssSetInputPortDataType(		      S, i, DYNAMICALLY_TYPED);
        ssSetInputPortComplexSignal(      S, i, COMPLEX_NO);   
		ssSetInputPortMatrixDimensions(   S, i, 1, 1);
    }

	// Sample Times (Only one used)
	ssSetNumSampleTimes(S, 1);

	// Let Simulink allocate P_WORK_LENGTH "Work Pointers" (initialized to NULL)
	ssSetNumPWork(S, P_WORK_LENGTH);
}


/**
 * mdlInitializeSampleTimes
 */
static void
mdlInitializeSampleTimes(SimStruct *S)
{
	WHERE_AM_I(S, "");

	// Check for correct number of sample times
	if (ssGetNumSampleTimes(S) != 1) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Wrong number of sample times registered. Expected 1, got %d.\n",
			ssGetNumSampleTimes(S)
		);
		ssSetErrorStatus(S, errorMessage);
		return;
	}

	// Fetch input parameters 'sample_time_factor' and 'decimation'
	double sampleTimeFactor = MatlabArrayHelper::getDoubleFromArray(
		ssGetSFcnParam(S, SAMPLE_TIME_FACTOR), "sample_time_factor", false, DBL_MIN
	);
	int decimation = MatlabArrayHelper::getIntFromArray(
		ssGetSFcnParam(S, DECIMATION), "decimation", false, 4, 64000000
	);

	// Get the size of the signal
	int portSize = ssGetOutputPortWidth(S, 0);

	// Calculate sample time (Oscillator frequency hardcoded for now...)
	double sampleTime;
	if (sampleTimeFactor > 0) {
		sampleTime = sampleTimeFactor * decimation * portSize / 64e6;
	} else if (sampleTimeFactor == -1) {
		sampleTime = -1;
	} else {
		_snprintf_s(errorMessage, _TRUNCATE,
			"'Sample Time Factor' of usrp_helper must be non-negative or -1. Got %f.\n",
			sampleTimeFactor
		);
		ssSetErrorStatus(S, errorMessage);
		return;
	}

	// Set sample time
	ssSetSampleTime(S, 0, sampleTime);
	ssSetOffsetTime(S, 0, 0.0);
}


/**
 * mdlStart
 */
#define MDL_START
static void
mdlStart(SimStruct *S)
{
	WHERE_AM_I(S, "");

	// Do nothing if this is a rapid accelerator or external mode simulation
	if (ssGetSimMode(S) == SS_SIMMODE_EXTERNAL) {
		return;
	}

	// Set options of this Block
	ssSetOptions(S,
		ssGetOptions(S) |
		// Needed since Simulink usually only calls mdlTerminate when mdlStart was run successfully.
		// However, we may create an USRP object in mdlStart, place it in PWork, and still call
		// ssSetErrorStatus afterwards.
		SS_OPTION_CALL_TERMINATE_ON_EXIT
	);

	// Fetch input parameters
	int usrpNumber = MatlabArrayHelper::getIntFromArray(
		ssGetSFcnParam(S, USRP_NUMBER), "usrp_number", false, 0, 1023
	);
	int decimation = MatlabArrayHelper::getIntFromArray(
		ssGetSFcnParam(S, DECIMATION), "decimation", false, 4, 64000000
	);
	int sampleWidth = UsrpHelper::sampleWidthString2SampleWidth(
		MatlabArrayHelper::getStringFromArray(
			ssGetSFcnParam(S, SAMPLE_WIDTH), "sample_width"
		)
	);
	int usbBlockSize = MatlabArrayHelper::getIntFromArray(
		ssGetSFcnParam(S, USB_BLOCK_SIZE), "usb_block_size",
		false, 0, fusb_sysconfig::max_block_size()
	);
	int usbNBlocks = MatlabArrayHelper::getIntFromArray(
		ssGetSFcnParam(S, USB_NBLOCKS), "usb_nblocks", false, 0
	);
	string firmware = MatlabArrayHelper::getStringFromArray(
		ssGetSFcnParam(S, FIRMWARE), "firmware"
	);
	string bitstream = MatlabArrayHelper::getStringFromArray(
		ssGetSFcnParam(S, BITSTREAM), "bitstream"
	);

	// Sanity check on usbBlockSize
	if (usbBlockSize % 512 != 0) {
		// Set it to the default value
		usbBlockSize = 0;
#ifdef MATLAB_MEX_FILE
		mexWarnMsgTxt(
			"'usb_block_size' must be a multiple of 512. "
			"Using the default value instead.\n"
		);
#endif
	}

	// Calculate number of used subdevs (and therefore used channels)
	vector <SimulinkUsrpHelper::subdevInfo> subdevInfo(MAX_SUBDEV_COUNT);
	int channelCount = SimulinkUsrpHelper::countEnabledSubdevices(
		S, MAX_SUBDEV_COUNT, SIDEA_SUBDEV1_DATATYPE, SUBDEV_PARAM_DISTANCE, subdevInfo
	);

	// Create an usrp_*_matlab object and store it in the work pointer vektor of the S-Function
	USRP_TYPE *usrp = NULL;
	try {
		usrp = new USRP_TYPE(
			usrpNumber, decimation, channelCount, -1, 0,
			usbBlockSize, usbNBlocks, bitstream, firmware
		);
		ssSetPWorkValue(S, USRP, usrp);
	} catch (std::runtime_error E) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Could not connect to USRP %d: %s\n",
			usrpNumber, E.what()
		);
		ssSetErrorStatus(S, errorMessage);
		return;
	}

	//// Set the usrp object into verbose mode
	//usrp->set_verbose(true);

	// Set sample width, if different from the default (2 Bytes)
	if (sampleWidth == 1) {
		usrp->usrp->set_format(usrp->usrp->make_format(8, 8));
		// TODO: set_usb_datarate(), so that uO gets displayed the correct number of times.
	}

	// Fetch all daugtherboard subdevices. This can be done without a conversion to weak pointers,
	// since the reference count will drop once we leave this function. (Currently only BasicRX and
	// LFRX have more than one subdev, see libursp, db_boards.cc, instantiate_dbs).
	vector <vector <db_base_sptr>> subdevs = usrp->usrp->db();
	db_base_sptr subdev;

	// Set frequency, gain etc. on each subdev
	int channelNum = 0;
	int mux = 0;
	for (int i = 0; i < MAX_SUBDEV_COUNT; i++) {
		if (!subdevInfo[i].enabled) {
			continue;
		}

		unsigned int subdevIndex = i % 2;
		unsigned int sideIndex = i / 2;

		// Hack to select subdev 3 (I&Q complex samples) on the BasicRX/LFRX
		string antenna;
		if (subdevIndex == 0) {
			antenna = MatlabArrayHelper::getStringFromArray(
				ssGetSFcnParam(S, SIDEA_SUBDEV1_ANTENNA + i * SUBDEV_PARAM_DISTANCE), "antenna"
			);
			if (antenna == "I&Q (BasicRX/LFRX only)") {
				subdevIndex += 2;
			}
		}

		// Get subdev db_base_sptr
		if (sideIndex >= subdevs.size() || subdevIndex >= subdevs[sideIndex].size()) {
			_snprintf_s(errorMessage, _TRUNCATE,
				"There is no subdevice %d on side %c of USRP %d.\n",
				subdevIndex, static_cast <char> ('A' + sideIndex), usrpNumber
			);
			ssSetErrorStatus(S, errorMessage);
			return;
		}
		subdev = subdevs[sideIndex][subdevIndex];

		// Tune the subdevice to the target frequency
		double frequency = MatlabArrayHelper::getDoubleFromArray(
			ssGetSFcnParam(S, SIDEA_SUBDEV1_FREQUENCY + i * SUBDEV_PARAM_DISTANCE),
			"frequency", false
		);
		
		// get subdev handle for dynamic frequency tuning
		if (usrp_source_dynfreq_subdevs[i] >= 0) {
			usrp_source_dynfreq_subdevs_ptr[i] = subdev;
		}

		usrp_tune_result tuneResult;
		if (!usrp->usrp->tune(channelNum, subdev, frequency, &tuneResult)) {
			_snprintf_s(errorMessage, _TRUNCATE,
				"Could not tune subdevice %d on side %c of USRP %d to the target "
				"frequency of %f Hz.\n",
				subdevIndex, static_cast <char> ('A' + sideIndex), usrpNumber, frequency
			);
			ssSetErrorStatus(S, errorMessage);
			return;
		}

		// Calculate values for the RX mux
		int muxForThisSubdev = usrp->usrp->determine_rx_mux_value(
			usrp_subdev_spec(sideIndex, subdevIndex)
		);
		muxForThisSubdev = (muxForThisSubdev & 0xff);
		mux |= (muxForThisSubdev << 8 * channelNum);

		// Set Gain
		float gain = MatlabArrayHelper::getFloatFromArray(
			ssGetSFcnParam(S, SIDEA_SUBDEV1_GAIN + i * SUBDEV_PARAM_DISTANCE),
			"gain", false
		);
		if(!subdev->set_gain(gain)) {
			_snprintf_s(errorMessage, _TRUNCATE,
				"Could not set the gain of subdevice %d on side %c of USRP %d to %f dB.\n",
				subdevIndex, static_cast <char> ('A' + sideIndex), usrpNumber, gain
			);
			ssSetErrorStatus(S, errorMessage);
			return;
		}

		// Select antenna
		if (subdevIndex == 0 && antenna != "Not applicable") {
			if (!subdev->select_rx_antenna(antenna)) {
				_snprintf_s(errorMessage, _TRUNCATE,
					"Could not set the rx antenna of subdevice %d on side %c "
					"of USRP %d to '%s'.\n",
					subdevIndex, static_cast <char> ('A' + sideIndex), usrpNumber,
					antenna.c_str()
				);
				ssSetErrorStatus(S, errorMessage);
				return;
			}
		}

		// Next channel
		channelNum++;
	}

	// Adjust MUX: Are we are operating on real samples?
	if ((mux & 0xf0) == 0xf0) {
		// Set the rest of the Q-Inputs of the mux to constant zero (f) as well.
		for (channelNum = channelCount; channelNum < MAX_CHANNEL_COUNT; channelNum++) {
			mux |= (0xf0 << 8 * channelNum);
		}
	}

	// Set the MUX value
	if (!usrp->usrp->set_mux(mux)) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Could not set the RX MUX value to '%08x'. "
			"Hint: Are you using real-valued and complex-valued sampling at the same time?\n",
			mux
		);
		ssSetErrorStatus(S, errorMessage);
		return;
	}

	// Determine the correct "copy" function for each port and store it in the PWork pointer
	// of this S-Function.
	int portNum = 0;
	for (channelNum = 0; channelNum < channelCount; channelNum++) {
		DTypeId dataType = ssGetOutputPortDataType(S, portNum);

		// Check that the correct number of ports was defined in mdlInitialzeSizes
		if (portNum >= ssGetNumOutputPorts(S)) {
			_snprintf_s(errorMessage, _TRUNCATE,
				"More channels than there are ports! "
				"This should not be possible to happen.\n"
			);
			ssSetErrorStatus(S, errorMessage);
			return;
		}

		// Check the complexity of the port
		if (ssGetOutputPortComplexSignal(S, portNum) == COMPLEX_YES) {
			// The variable copyFcnPointer is a pointer to a function pointer. This is done because
			// it is impossible to cast pointers to class member functions to void*
			USRP_TYPE::copyFunctionPointerComplex *copyFcnPointer;
			copyFcnPointer = new USRP_TYPE::copyFunctionPointerComplex;

			// Assign the correct function to each function pointer
			if (dataType == SS_DOUBLE) {
				*copyFcnPointer = &USRP_TYPE::copySimulinkSignal <CREAL_T>;
			} else if (dataType == SS_SINGLE) {
				*copyFcnPointer = &USRP_TYPE::copySimulinkSignal <CREAL32_T>;
			} else if (dataType == SS_INT8) {
				*copyFcnPointer = &USRP_TYPE::copySimulinkSignal <CINT8_T>;
			} else if (dataType == SS_INT16) {
				*copyFcnPointer = &USRP_TYPE::copySimulinkSignal <CINT16_T>;
			} else if (dataType == SS_INT32) {
				*copyFcnPointer = &USRP_TYPE::copySimulinkSignal <CINT32_T>;
			} else {
				_snprintf_s(errorMessage, _TRUNCATE, "Unsupported datatype for port %d.\n",
					portNum
				);
				ssSetErrorStatus(S, errorMessage);
				return;
			}

			// Save the address of the pointer
			ssSetPWorkValue(S, COPY_FUNCTION_POINTER0 + channelNum, copyFcnPointer);

			// Next port
			portNum++;
		} else {
			// The variable copyFcnPointer is a pointer to a function pointer. This is done because
			// it is impossible to cast pointers to class member functions to void*
			USRP_TYPE::copyFunctionPointerReal *copyFcnPointer;
			copyFcnPointer = new USRP_TYPE::copyFunctionPointerReal;

			// Assign the correct function to each function pointer
			if (dataType == SS_DOUBLE) {
				*copyFcnPointer = &USRP_TYPE::copySimulinkSignal <REAL_T>;
			} else if (dataType == SS_SINGLE) {
				*copyFcnPointer = &USRP_TYPE::copySimulinkSignal <REAL32_T>;
			} else if (dataType == SS_INT8) {
				*copyFcnPointer = &USRP_TYPE::copySimulinkSignal <INT8_T>;
			} else if (dataType == SS_INT16) {
				*copyFcnPointer = &USRP_TYPE::copySimulinkSignal <INT16_T>;
			} else if (dataType == SS_INT32) {
				*copyFcnPointer = &USRP_TYPE::copySimulinkSignal <INT32_T>;
			} else {
				_snprintf_s(errorMessage, _TRUNCATE, "Unsupported datatype for ports %d and %d.\n",
					portNum, portNum + 1
				);
				ssSetErrorStatus(S, errorMessage);
				return;
			}

			// Save the address of the pointer
			ssSetPWorkValue(S, COPY_FUNCTION_POINTER0 + channelNum, copyFcnPointer);

			// Next port
			portNum += 2;
		}
	}

	// Allocate the buffer of the USRP
	if (!usrp->allocateBuffer(ssGetOutputPortWidth(S, 0))) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Could not allocate memory for the local buffer of the USRP %d.\n",
			usrpNumber
		);
		ssSetErrorStatus(S, errorMessage);
		return;
	}

	// Start data streaming
	if (!usrp->usrp->start()) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Could not start streaming of data from USRP %d.\n",
			usrpNumber
		);
		ssSetErrorStatus(S, errorMessage);
		return;
	}
}


/**
 * mdlOutputs
 */
static void
mdlOutputs(SimStruct *S, int_T tid)
{
	// Security measure: do we really have an USRP object created before?
	if (ssGetPWork(S) == NULL || ssGetPWorkValue(S, USRP) == NULL) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"No USRP object has been initialized before. This should not be possible to happen.\n"
		);
		ssSetErrorStatus(S, errorMessage);
		return;
	}
	USRP_TYPE *usrp = static_cast <USRP_TYPE *> (ssGetPWorkValue(S, USRP));

	// Fetch data from the USRP
	if (!usrp->transfer()) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Reading from the USRP failed.\n"
		);
		ssSetErrorStatus(S, errorMessage);
		return;
	}

	// Copy the data between each Simulink signal and the USRP's buffer
	bool result = true;
	int portNum = 0;
	for (int channelNum = 0; channelNum < usrp->usrp->nchannels(); channelNum++) {
		void *copyFcnVoidPointer = ssGetPWorkValue(S, COPY_FUNCTION_POINTER0 + channelNum);
		if (copyFcnVoidPointer == NULL) {
			_snprintf_s(errorMessage, _TRUNCATE,
				"No function has been set for port %d. "
				"This should not be possible to happen.\n",
				portNum
			);
			ssSetErrorStatus(S, errorMessage);
			return;
		}

		// Check the complexity of the port
		if (ssGetOutputPortComplexSignal(S, portNum) == COMPLEX_YES) {
			USRP_TYPE::copyFunctionPointerComplex *copyFcnPointer =
				static_cast <USRP_TYPE::copyFunctionPointerComplex *> (copyFcnVoidPointer)
			;

			// Call the port function
			result &= (usrp->**copyFcnPointer) (
				ssGetOutputPortWidth(S, 0), channelNum, ssGetOutputPortSignal(S, portNum)
			);

			// Next port
			portNum++;
		} else {
			USRP_TYPE::copyFunctionPointerReal *copyFcnPointer =
				static_cast <USRP_TYPE::copyFunctionPointerReal *> (copyFcnVoidPointer)
			;

			// Call the port function, using the current and the next port,
			// and incrementing portNum
			result &= (usrp->**copyFcnPointer) (
				ssGetOutputPortWidth(S, 0), channelNum,
				ssGetOutputPortSignal(S, portNum), ssGetOutputPortSignal(S, portNum+1)
			);

			// Next port
			portNum += 2;
		}
	}

	// Dynamic frequency tuning
	int channelNum = 0;
	for (int i = 0; i < MAX_SUBDEV_COUNT; i++) {
			
			// If dynamic frequency tuning for subdevice not selectet --> exit
			if (usrp_source_dynfreq_subdevs[i] < 0) {
				continue;
			}
			
			// Allocate variables
			int dynfreq_uPtrs_width;
			InputRealPtrsType dynfreq_uPtrs;
		
			// get frequency input port value
			dynfreq_uPtrs = ssGetInputPortRealSignalPtrs(S,usrp_source_dynfreq_subdevs[i]);
			dynfreq_uPtrs_width = ssGetInputPortWidth(S,usrp_source_dynfreq_subdevs[i]);
			
			// Don't tune unnecessarily
			if (usrp_source_dynfreq_subdevs_tune_hold[i] == *dynfreq_uPtrs[0]) {
				continue;
			}

			// Tune frequency
			usrp_tune_result tuneResult;
			if (!usrp->usrp->tune(channelNum, usrp_source_dynfreq_subdevs_ptr[i], *dynfreq_uPtrs[0], &tuneResult)) {

				// Get board information
				unsigned int sideIndex = i / 2;
				int usrpNumber = MatlabArrayHelper::getIntFromArray(
					ssGetSFcnParam(S, USRP_NUMBER), "usrp_number", false, 0, 1023
					);

				// Print error message if frequency not tuneable
				_snprintf_s(errorMessage, _TRUNCATE,
					"Could not tune subdevice %d on side %c of USRP %d to the target "
					"frequency of %f Hz.\n",
					i, static_cast <char> ('A' + sideIndex), usrpNumber, *dynfreq_uPtrs[0]
					);
				
				ssSetErrorStatus(S, errorMessage);
				
				return;
			}
			
			// Save the tuned frequency
			usrp_source_dynfreq_subdevs_tune_hold[i] = *dynfreq_uPtrs[0];

	}


	// Did any one of the functions not succeed? Then abort the simulation, so that
	// Simulink does not look into the signal.
	if (!result) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Extracting a signal from the read USRP buffer failed.\n"
		);
		ssSetErrorStatus(S, errorMessage);
		return;
	}
}


/**
 * mdlTerminate
 */
static void
mdlTerminate(SimStruct *S)
{
	WHERE_AM_I(S, "");

	// Do we have to clean up at all?
	if (ssGetPWork(S) == NULL) {
		return;
	}

	// If we have created an USRP object, lets destoy it now
	if (ssGetPWorkValue(S, USRP) != NULL) {
		USRP_TYPE *usrp = NULL;
		usrp = static_cast <USRP_TYPE *> (ssGetPWorkValue(S, USRP));
		usrp->usrp->stop();
		delete usrp;
	}

	// Destroy every allocated function pointer
	for (int i = COPY_FUNCTION_POINTER0; i <= COPY_FUNCTION_POINTER_COUNT; i++) {
		if (ssGetPWorkValue(S, i) != 0) {
			delete ssGetPWorkValue(S, i);
		}
	}
}


/**************************************************************************************************
 * Required S-function trailer
 **************************************************************************************************/
/*
 * This file is a C++ file, but it must be compiled with the C calling conventions
 * so that the Matlab/Simulink engine is able to call this file
 */
extern "C"
{

#ifdef  MATLAB_MEX_FILE    /* Is this file being compiled as a MEX-file? */
#include "simulink.c"      /* MEX-file interface mechanism */
#else
#include "cg_sfun.h"       /* Code generation registration function */
#endif

}
