/**
 * Helper functions to handle parameter extraction from mxArrays
 *
 * @version:	$Id: MatlabArrayHelper.cpp 220 2009-05-15 07:39:45Z nagel $
 * @author:		Enno Klasing <klasing@int.uni-karlsruhe.de>
 * @copyright:	Institut fr Nachrichtentechnik, http://www.int.uni-karlsruhe.de
 * @license:	GNU General Public License, Version 3
 */

#include "matlabArrayHelper.h"
#ifdef MATLAB_MEX_FILE
#include "mex.h"
#else
#include <rt_matrx.h>
#endif

static char errorMessage[256];


double
MatlabArrayHelper::getDoubleFromArray(
	const mxArray *inArray, const std::string paramName, bool strict, double rangeMin, double rangeMax
) {
#ifdef MATLAB_MEX_FILE
	if (mxGetNumberOfElements(inArray) != 1) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Parameter '%s' must be a scalar, not an array. The parameter has %d elements.\n",
			paramName.c_str(),
			mxGetNumberOfElements(inArray)
		);
		mexErrMsgTxt(errorMessage);
	}

	if (mxIsComplex(inArray)) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Parameter '%s' must not be complex.\n",
			paramName.c_str()
		);
		mexErrMsgTxt(errorMessage);
	}

	if (strict && !mxIsDouble(inArray) || !mxIsNumeric(inArray)) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Parameter '%s' must be %s, '%s' given.\n",
			paramName.c_str(),
			strict ? "numeric" : "of type 'double'",
			mxGetClassName(inArray)
		);
		mexErrMsgTxt(errorMessage);
	}
#endif

	double result = mxGetPr(inArray)[0];
	if (result < rangeMin || result > rangeMax) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Parameter '%s' is out of range. Must be between %f and %f.\n",
			paramName.c_str(), rangeMin, rangeMax
		);
#ifdef MATLAB_MEX_FILE
		mexErrMsgTxt(errorMessage);
#else
		result = rangeMin;
		printf("%s", errorMessage);
#endif
	}

	return result;
}

int
MatlabArrayHelper::getIntFromArray(
	const mxArray *inArray, const std::string paramName, bool strict, int rangeMin, int rangeMax
) {
#ifdef MATLAB_MEX_FILE
	if (mxGetNumberOfElements(inArray) != 1) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Parameter '%s' must be a scalar, not an array. The parameter has %d elements.\n",
			paramName.c_str(),
			mxGetNumberOfElements(inArray)
		);
		mexErrMsgTxt(errorMessage);
	}

	if (mxIsComplex(inArray)) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Parameter '%s' must not be complex.\n",
			paramName.c_str()
		);
		mexErrMsgTxt(errorMessage);
	}

	if (strict && !(
			mxIsInt8(inArray) || mxIsInt16(inArray) || mxIsInt32(inArray)
		) || !mxIsNumeric(inArray)
	) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Parameter '%s' must be %s, '%s' given.\n",
			paramName.c_str(),
			strict ? "numeric" : "of type 'double'",
			mxGetClassName(inArray)
		);
		mexErrMsgTxt(errorMessage);
	}
#endif

	double temp = mxGetPr(inArray)[0];
	int result = static_cast <int> (temp);
	if (temp < INT_MIN || temp > INT_MAX ||
		result < rangeMin || result > rangeMax
	) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Parameter '%s' is out of range. Must be between %d and %d.\n",
			paramName.c_str(), rangeMin, rangeMax
		);
#ifdef MATLAB_MEX_FILE
		mexErrMsgTxt(errorMessage);
#else
		result = rangeMin;
		printf("%s", errorMessage);
#endif
	}

	return result;
}

bool
MatlabArrayHelper::getBoolFromArray(
	const mxArray *inArray, const std::string paramName, bool strict
) {
#ifdef MATLAB_MEX_FILE
	if (mxGetNumberOfElements(inArray) != 1) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Parameter '%s' must be a scalar, not an array. The parameter has %d elements.\n",
			paramName.c_str(),
			mxGetNumberOfElements(inArray)
		);
		mexErrMsgTxt(errorMessage);
	}

	if (mxIsComplex(inArray)) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Parameter '%s' must not be complex.\n",
			paramName.c_str()
		);
		mexErrMsgTxt(errorMessage);
	}

	if (strict && !mxIsLogical(inArray) || !mxIsNumeric(inArray)) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Parameter '%s' must be %s, '%s' given.\n",
			paramName.c_str(),
			strict ? "numeric" : "of type 'logical'",
			mxGetClassName(inArray)
		);
		mexErrMsgTxt(errorMessage);
	}
#endif

	return mxGetPr(inArray)[0] != 0;
}

#ifdef MATLAB_MEX_FILE
std::wstring
MatlabArrayHelper::getWStringFromArray(
	const mxArray *inArray, const std::string paramName
) {
	if (!mxIsChar(inArray)) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Parameter '%s' must be of type 'char', '%s' given.\n",
			paramName.c_str(),
			mxGetClassName(inArray)
		);
		mexErrMsgTxt(errorMessage);
	}

	return std::wstring(mxGetChars(inArray), mxGetN(inArray));
}
#endif

std::string
MatlabArrayHelper::getStringFromArray(
	const mxArray *inArray, const std::string paramName
) {
	std::string result;

#ifdef MATLAB_MEX_FILE
	if (!mxIsChar(inArray)) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Parameter '%s' must be of type 'char', '%s' given.\n",
			paramName.c_str(),
			mxGetClassName(inArray)
		);
		mexErrMsgTxt(errorMessage);
	}
#endif

	char *cResult = new char[mxGetN(inArray) + 1];
	mxGetString(inArray, cResult, mxGetN(inArray) + 1);
	result = std::string(cResult, mxGetN(inArray));
	delete[] cResult;

	return result;
}
