/**
 * Helper functions to call various Matlab functions
 *
 * @version:	$Id: MatlabFunctionHelper.cpp 220 2009-05-15 07:39:45Z nagel $
 * @author:		Enno Klasing <klasing@int.uni-karlsruhe.de>
 * @copyright:	Institut fr Nachrichtentechnik, http://www.int.uni-karlsruhe.de
 * @license:	GNU General Public License, Version 3
 */

#include "MatlabFunctionHelper.h"
#include "MatlabArrayHelper.h"

using namespace std;

static char errorMessage[256];


/**
 * Executes "paramValue = get_param(object, paramName)" in Matlab.
 */
mxArray*
MatlabFunctionHelper::getParam(
	const mxArray *object, const std::string paramName
) {
	bool result = false;
	mxArray *paramValue = NULL;

	const int nrhs = 2;
	mxArray *prhs[nrhs];

	prhs[0] = const_cast <mxArray *> (object);
	prhs[1] = mxCreateString(paramName.c_str());

	result = mexCallMATLAB(1, &paramValue, nrhs, prhs, "get_param") == 0;

	mxDestroyArray(prhs[1]);

	if (!result) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Error while fetching the value of parameter '%s'.\n",
			paramName.c_str()
		);
		mexErrMsgTxt(errorMessage);
	}

	return paramValue;
}


/**
 * Executes "set_param(object, paramName, paramValue)" in Matlab.
 */
void
MatlabFunctionHelper::setParam(
	const mxArray *object, const std::string paramName, const mxArray *paramValue
) {
	bool result = false;
	const int nrhs = 3;
	mxArray *prhs[nrhs];

	prhs[0] = const_cast <mxArray *> (object);
	prhs[1] = mxCreateString(paramName.c_str());
	prhs[2] = const_cast <mxArray *> (paramValue);

	result = mexCallMATLAB(0, NULL, nrhs, prhs, "set_param") == 0;

	mxDestroyArray(prhs[1]);

	if (!result) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Error while setting the value of parameter '%s'.\n",
			paramName.c_str()
		);
		mexErrMsgTxt(errorMessage);
	}
}


/**
 * Executes "set_param(object, params[0], params[1], ...)" in Matlab.
 */
void
MatlabFunctionHelper::setParam(const mxArray *object, const vector <string> params)
{
	bool result = false;

	// Check length of params
	if (params.size() % 2 != 0 || params.size() < 2) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Wrong number of parameters given.\n"
		);
		mexErrMsgTxt(errorMessage);
	}

	// Check if there are parameters which have not changed. If every param has not changed,
	// we can save a call to mdlInitializeSizes of the corresponding object by not calling
	// set_param at all.
	bool allParamsIdentical = true;
	for (unsigned int i = 0; i < params.size(); i += 2) {
		mxArray *paramValue = getParam(object, params[i]);
		string paramValueString = MatlabArrayHelper::getStringFromArray(paramValue, params[i]);
		mxDestroyArray(paramValue);

		if (paramValueString != params[i+1]) {
			allParamsIdentical = false;
			break;
		}
	}
	if (allParamsIdentical) {
		return;
	}

	// Create mxArray[]
	const int nrhs = params.size() + 1;
	mxArray **prhs = new mxArray*[nrhs];

	// Populate the mxArrays
	prhs[0] = const_cast <mxArray *> (object);
	for (int i = 1; i < nrhs; i++) {
		prhs[i] = mxCreateString(params[i-1].c_str());
	}

	// Call Matlab
	result = mexCallMATLAB(0, NULL, nrhs, prhs, "set_param") == 0;

	// Clean up
	for (int i = 1; i < nrhs; i++) {
		mxDestroyArray(prhs[i]);
	}
	delete[] prhs;

	if (!result) {
		_snprintf_s(errorMessage, _TRUNCATE,
			"Error while setting the value of parameter '%s'.\n",
			params[0].c_str()
		);
		mexErrMsgTxt(errorMessage);
	}
}


/**
 * Executes "currentBlock = gcb()" in Matlab.
 */
mxArray*
MatlabFunctionHelper::gcb()
{
	mxArray* currentBlock = NULL;

	if (mexCallMATLAB(1, &currentBlock, 0, NULL, "gcb") != 0) {
		mexErrMsgTxt(
			"Error while fetching the name of the current block. "
			"This function can only be used from within a Simulink Callback Function.\n"
		);
	}

	return currentBlock;
}


/**
 * Executes "currentModel = bdroot()" in Matlab.
 */
mxArray*
MatlabFunctionHelper::bdroot()
{
	mxArray* currentBlock = NULL;

	if (mexCallMATLAB(1, &currentBlock, 0, NULL, "bdroot") != 0) {
		mexErrMsgTxt(
			"Error while fetching the name of the current system (bdroot). "
			"This function can only be used from within a Simulink Callback Function.\n"
		);
	}

	return currentBlock;
}

	
/**
 * Executes "param = str2num(param)" in Matlab.
 */
void
MatlabFunctionHelper::str2Num(mxArray **param)
{
	mxArray *result = NULL;

	if (mexCallMATLAB(1, &result, 1, param, "str2num") != 0) {
		mexErrMsgTxt("Error while executing str2num().\n");
	}

	mxDestroyArray(*param);
	*param = result;
}
