/**
 * Helper functions to call various Matlab functions
 *
 * @version:	$Id: MatlabFunctionHelper.h 220 2009-05-15 07:39:45Z nagel $
 * @author:		Enno Klasing <klasing@int.uni-karlsruhe.de>
 * @copyright:	Institut fr Nachrichtentechnik, http://www.int.uni-karlsruhe.de
 * @license:	GNU General Public License, Version 3
 */

#include <simstruc.h>
#include <string>
#include <vector>


namespace MatlabFunctionHelper
{
	/**
	 * Executes "paramValue = get_param(object, paramName)" in Matlab.
	 *
	 * @param	mxArray		Object on which getParam shall be executed, e.g. a Simulink block.
	 * @param	string		Name of the parameter.
	 * @return	mxArray		The parameter value. It is your responsibility to free this array
	 *						prior to calling this function in case something was assigned to it.
	 *						Also, it is your responsibility to free this array when you are done
	 *						using the result.
	 * @throws
	 */
	mxArray* getParam(const mxArray *object, const std::string paramName);

	/**
	 * Executes "set_param(object, paramName, paramValue)" in Matlab.
	 *
	 * @param	mxArray		Object on which getParam shall be executed, e.g. a Simulink block.
	 * @param	string		Name of the parameter.
	 * @param	mxArray		Value of the parameter.
	 * @throws
	 */
	void setParam(const mxArray *object, const std::string paramName, const mxArray *paramValue);

	/**
	 * Executes "set_param(object, params[0], params[1], ...)" in Matlab.
	 *
	 * params must have an even length, only string parameters can be set with this function.
	 *
	 * @param	mxArray		Object on which getParam shall be executed, e.g. a Simulink block.
	 * @param	vector<string>
	 *						Alternating name/value pairs of the parameters.
	 * @throws
	 */
	void setParam(const mxArray *object, const std::vector <std::string> params);

	/**
	 * Executes "currentBlock = gcb()" in Matlab.
	 *
	 * @return	mxArray		The path to the current block. It is your responsibility to free this
	 *						array prior to calling this function in case something was assigned to
	 *						it. Also, it is your responsibility to clear this array when you are
	 *						done using the result.
	 * @throws
	 */
	mxArray* gcb();

	/**
	 * Executes "currentModel = bdroot()" in Matlab.
	 *
	 * @return	mxArray		The path to the current block. It is your responsibility to free this
	 *						array prior to calling this function in case something was assigned to
	 *						it. Also, it is your responsibility to clear this array when you are
	 *						done using the result.
	 * @throws
	 */
	mxArray* bdroot();

	/**
	 * Executes "param = str2num(param)" in Matlab.
	 *
	 * @param	mxArray		String to be converted. This array will be deleted and replaced with the
	 *						result when this function finishes.
	 * @throws
	 */
	void str2Num(mxArray **param);
};
