/**
 * Helper functions to find USRP blocks in a Simulink System
 *
 * @version:	$Id: MatlabUsrpHelper.cpp 220 2009-05-15 07:39:45Z nagel $
 * @author:		Enno Klasing <klasing@int.uni-karlsruhe.de>
 * @copyright:	Institut fr Nachrichtentechnik, http://www.int.uni-karlsruhe.de
 * @license:	GNU General Public License, Version 3
 */

#include "MatlabUsrpHelper.h"
#include "MatlabArrayHelper.h"
#include "MatlabFunctionHelper.h"
#include <fusb.h>
#include "mex.h"

using namespace std;

/**
 * Returns a vector of paths to all usrp_helper blocks in the specified system
 */
vector <string>
MatlabUsrpHelper::findUsrpHelperBlocks(const mxArray *system)
{
	const int nrhs = 5;
	mxArray *prhs[nrhs];
	mxArray *blockNames;
	vector <string> result;

	// Call "find_system(system, 'BlockType', 'S-Function', 'FunctionName', 'usrp_helper')"
	prhs[0] = const_cast <mxArray *> (system);
	prhs[1] = mxCreateString("BlockType");
	prhs[2] = mxCreateString("S-Function");
	prhs[3] = mxCreateString("FunctionName");
	prhs[4] = mxCreateString("usrp_helper");
	if (mexCallMATLAB(1, &blockNames, nrhs, prhs, "find_system") != 0) {
		mexErrMsgTxt("Error while executing find_system().\n");
	}
	for (int i = 1; i < nrhs; i++) { // Don't clear "system"
		mxDestroyArray(prhs[i]);
	}

	// Extract block names from the returned mxArray
	if (mxIsCell(blockNames)) {
		for (unsigned int i = 0; i < mxGetNumberOfElements(blockNames); i++) {
			mxArray *blockName = mxGetCell(blockNames, i);
			if (blockName && mxIsChar(blockName)) {
				char *cBlockName = mxArrayToString(blockName);
				result.push_back(string(cBlockName, mxGetN(blockName)));
				mxFree(cBlockName);
			}
		}
	}

	// Free returned mxArray
	mxDestroyArray(blockNames);

	return result;
}

/**
 * Returns the paths to a usrp_helper block which controls parameters for the Nth
 * USRP on this computer.
 */
string
MatlabUsrpHelper::findUsrpHelperBlock(const mxArray* system, const mxArray* usrpNumber)
{
	string result;

	int usrpNumberRequested = MatlabArrayHelper::getIntFromArray(
		usrpNumber, "usrp_number", false, 0, 1023
	);

	vector <string> blocks = findUsrpHelperBlocks(system);
	vector <string>::const_iterator blocksIter;

	for(blocksIter = blocks.begin(); blocksIter != blocks.end(); blocksIter++) {
		// Extract parameter usrp_number from the block
		mxArray *block = mxCreateString(blocksIter->c_str());
		mxArray *usrpNumberArray = MatlabFunctionHelper::getParam(block, "usrp_number");
		mxDestroyArray(block);
		MatlabFunctionHelper::str2Num(&usrpNumberArray);
		int usrpNumberActual = MatlabArrayHelper::getIntFromArray(
			usrpNumberArray, "usrp_number", false, 0, 1023
		);
		mxDestroyArray(usrpNumberArray);

		// Compare it with the requested usrp_number
		if (usrpNumberActual == usrpNumberRequested) {
			if (result.empty()) {
				result = *blocksIter;
			} else {
				mexErrMsgTxt("There is more than one 'usrp_helper' block configured for the same USRP.\n");
			}
		}
	}

	return result;
}

/**
 * Fetches a parameter from a usrp_helper block.
 */
mxArray *
MatlabUsrpHelper::fetchParamFromHelperBlock(
	const mxArray *system, const mxArray *usrpNumber, const string paramName
) {
	mxArray *result = NULL;

	// Fetch the usrp_helper block for this USRP
	string usrpHelperBlock = MatlabUsrpHelper::findUsrpHelperBlock(system, usrpNumber);

	if (!usrpHelperBlock.empty()) {
		// Fetch the parameter from the helper block
		mxArray *block = mxCreateString(usrpHelperBlock.c_str());
		result = MatlabFunctionHelper::getParam(block, paramName);
		mxDestroyArray(block);
	}

	return result;
}
