/**
 * Helper file for fetching information from the USRP hardware
 *
 * @version:	$Id: SimulinkUsrpHelper.cpp 220 2009-05-15 07:39:45Z nagel $
 * @author:		Enno Klasing <klasing@int.uni-karlsruhe.de>
 * @copyright:	Institut fr Nachrichtentechnik, http://www.int.uni-karlsruhe.de
 * @license:	GNU General Public License, Version 3
 */

#include "SimulinkUsrpHelper.h"
#include "MatlabArrayHelper.h"
#include <Windows.h>

using namespace std;


/**
 * Standard constructor for struct SimulinkUsrpHelper::subdevInfo
 */
SimulinkUsrpHelper::subdevInfo::subdevInfo()
:
	enabled(false),
	complexOutput(COMPLEX_NO)
{
}


/**
 * Counts the number of enabled subdevices
 */
int
SimulinkUsrpHelper::countEnabledSubdevices(
	const SimStruct *S,
	const int MAX_SUBDEV_COUNT, const int PARAM_DATATYPE_INDEX, const int PARAM_DISTANCE,
	vector <subdevInfo> &subdeviceInfo, int *outputCount
) {
	int subdevCount = 0;
	subdevInfo subdev;

	// Check number of S-Function parameters
	if (ssGetNumSFcnParams(S) < PARAM_DATATYPE_INDEX + MAX_SUBDEV_COUNT * PARAM_DISTANCE) {
#ifdef MATLAB_MEX_FILE
		mexErrMsgTxt(
			"SimulinkUsrpHelper::countEnabledSubdevices: To few S-Function parameters.\n"
		);
#else
		printf(
			"SimulinkUsrpHelper::countEnabledSubdevices: To few S-Function parameters.\n"
		);
		return 0;
#endif
	}

	// Iterate over all subdev parameters
	for (int i = 0; i < MAX_SUBDEV_COUNT; i++) {
		// Fetch input parameter 'data_type'
		subdev.dataType = MatlabArrayHelper::getStringFromArray(
			ssGetSFcnParam(S, PARAM_DATATYPE_INDEX + i * PARAM_DISTANCE), "sideX_subdevX_datatype"
		);

		// Check desired complexity of the output port signal
		wstring::size_type firstOccurence = subdev.dataType.find(" Complex");
		if (firstOccurence != wstring::npos) {
			subdev.complexOutput = COMPLEX_YES;
			subdev.dataType.erase(firstOccurence, wstring::npos);
		} else {
			subdev.complexOutput = COMPLEX_NO;
		}

		// Add this subdev to the number of enabled subdevs
		subdev.enabled = subdev.dataType != "Disabled";
		if (subdev.enabled) {
			subdevCount++;
			if (outputCount) {
				*outputCount += subdev.complexOutput == COMPLEX_NO ? 2 : 1;
			}
		}

		// Add the information to the subdevInfo
		if (subdeviceInfo.size() == MAX_SUBDEV_COUNT) {
			subdeviceInfo[i] = subdev;
		}
	}

	return subdevCount;
}

/**
 * Converts a data-type string into the number of a Simulink builtin datatype
 */
bool
SimulinkUsrpHelper::dataTypeString2BuiltinDataType(
	const subdevInfo subdev, BuiltInDTypeId &dataType
) {
	// Check desired output data type
	if (subdev.dataType == "INT8") {
		dataType = SS_INT8;
	} else if (subdev.dataType == "INT16") {
		dataType = SS_INT16;
	} else if (subdev.dataType == "INT32") {
		dataType = SS_INT32;
	} else if (subdev.dataType == "SINGLE") {
		dataType = SS_SINGLE;
	} else if (subdev.dataType == "DOUBLE") {
		dataType = SS_DOUBLE;
	} else {
		return false;
	}

	return true;
}


/**
 * Sets the environment Variable USRP_PATH, which is then used by find_file in usrp_prims.cc
 */
bool
SimulinkUsrpHelper::setPathEnvironmentVar()
{
	static bool firstUse = true;

	if (!firstUse) {
		return true;
	}

	const char *moduleNames[] = {
		"usrp_helper.mexw32",
		"usrp_helper.mexw64",
		"usrp_helper.mexw",
		"usrp_sink.mexw32",
		"usrp_sink.mexw64",
		"usrp_sink.mexw",
		"usrp_source.mexw32",
		"usrp_source.mexw64",
		"usrp_source.mexw",
		0
	};

	// Iterate over all possible DLL names, and find a loaded one
	const char **moduleName = moduleNames;
	HMODULE module = NULL;
	while (moduleName) {
		module = GetModuleHandleA(*moduleName);
		if (module != NULL) {
			break;
		}
		moduleName++;
	}

	// Have we found a module?
	if (module == NULL) {
		printf(
			"Could not find the DLL of the current Mex-Function. You probably need to set the "
			"environment variable USRP_PATH by yourself.\n"
		);
		return false;
	}

	// Find the filename of the module
	char filename[1024] = "";
	DWORD result = GetModuleFileNameA(module, filename, sizeof(filename));
	if (result == 0 || result == sizeof(filename)) {
		printf(
			"Could not find the name of the current Mex-Function. You probably need to set the "
			"environment variable USRP_PATH by yourself.\n"
		);
		return false;
	}

	// Extract the path
	char *end = strrchr(filename, '\\');
	if (end == filename) {
		printf(
			"Could not extract the path of the current Mex-Function. You probably need to set the "
			"environment variable USRP_PATH by yourself.\n"
		);
		return false;
	}
	*end = 0;
	strcat_s(filename, "\\..\\firmware");

	// Set the environment variable
	if (_putenv_s("USRP_PATH", filename) != 0) {
		printf(
			"Could not set the environment variable USRP_PATH to '%s	'.\n", filename
		);
		return false;
	}

	firstUse = false;
	return true;
}