/**
 * Helper file for fetching information from the USRP hardware
 *
 * @version:	$Id: UsrpHelper.cpp 220 2009-05-15 07:39:45Z nagel $
 * @author:		Enno Klasing <klasing@int.uni-karlsruhe.de>
 * @copyright:	Institut fr Nachrichtentechnik, http://www.int.uni-karlsruhe.de
 * @license:	GNU General Public License, Version 3
 */

#include "UsrpHelper.h"

#include "UsrpSimulinkSource.h"
#include "UsrpSimulinkSink.h"
#include <usrp_prims.h>
#include <sstream>

using namespace std;
using namespace UsrpHelper;


/**
 * Formats a frequency into a human-readable form.
 *
 * This is a private local function.
 */
string
freq2String(double freq)
{
	ostringstream result;
	//result << showpoint;

	if (freq > 1e9 || freq < -1e9) {
		result << freq/1e9 << " GHz";
	} else if (freq > 1e6 || freq < -1e6) {
		result << freq/1e6 << " MHz";
	} else if (freq > 1e3 || freq < -1e3) {
		result << freq/1e3 << " kHz";
	} else {
		result << freq << " Hz";
	}

	return result.str();
}


/**
 * Extracts gain range and frequency range from a db_base_sptr and puts the result into
 * a UsrpHelper::subdevice.
 *
 * This is a private local function.
 */
void
fetchDataFromUsrpDb(db_base_sptr db, subdevice &subdev)
{
	// Subdev name
	subdev.name = db->name();

	// Subdev gain range
	ostringstream gainRange;
	gainRange <<
		"Min: " << db->gain_min() << " dB, " <<
		"Max: " << db->gain_max() << " dB, " <<
		"Step Size: "  << db->gain_db_per_step() << " dB."
	;
	subdev.gainRange = gainRange.str();

	// Subdev frequency range
	ostringstream frequencyRange;
	frequencyRange <<
		"Min: " << freq2String(db->freq_min())  << ", " <<
		"Max: " << freq2String(db->freq_max()) << "."
	;
	subdev.frequencyRange = frequencyRange.str();
}


/**
 * Standard constructor for struct UsrpHelper::subdevice
 */
UsrpHelper::subdevice::subdevice ()
:
	name(""), gainRange("Unknown."), frequencyRange("Unknown.")
{
}


/**
 * Returns a string with the serial number of the Nth USRP Board on this computer.
 */
string
UsrpHelper::boardInfo(int boardNumber)
{
	ostringstream result;

	// Validate range of the board number
	if (boardNumber < 0 || boardNumber > 1023) {
		result << "Unknown (Board Number #" << boardNumber << " out of range).";
		return result.str();
	}

	// Create an UsrpSimulinkSource object
	UsrpSimulinkSource *usrp = NULL;
	try {
		usrp = new UsrpSimulinkSource(boardNumber);
	} catch (std::runtime_error E) {
		result << "Unknown (USRP Board #" << boardNumber << " not found).";
	}

	/*
	 * Fetch data from the UsrpSimulinkSource object.
	 * This is done outside of the try/catch block to catch only errors during object instanciation
	 */
	if (usrp) {
		result << "S/N: " << usrp->usrp->serial_number() << ".\n";
	}

	// Destroy the UsrpSimulinkSource object
	if (usrp) {
		delete usrp;
	}

	return result.str();
}


/**
 * Returns a string with the names of all daughterboards on the Nth USRP Board on this computer.
 */
void
UsrpHelper::daughterboardInfo(int boardNumber, string &rxDaughterboards, string &txDaughterboards)
{
	ostringstream result;

	// Validate range of the board number
	if (boardNumber < 0 || boardNumber > 1023) {
		result << "Unknown (Board Number #" << boardNumber << " out of range).";
		rxDaughterboards = result.str();
		txDaughterboards = result.str();
		return;
	}

	// Create an UsrpSimulinkSource object
	UsrpSimulinkSource *usrpSource = NULL;
	UsrpSimulinkSink *usrpSink = NULL;
	try {
		usrpSource = new UsrpSimulinkSource(boardNumber);
		usrpSink = new UsrpSimulinkSink(boardNumber);
	} catch (std::runtime_error E) {
		result << "Unknown (USRP Board #" << boardNumber << " not found).";
		rxDaughterboards = result.str();
		txDaughterboards = result.str();
		return;
	}

	/*
	 * Fetch data from the usrp_*_matlab objects.
	 * This is done outside of the try/catch block to catch only errors during object instanciation
	 */
	if (usrpSource) {
		result << "RX A: " << usrp_dbid_to_string(usrpSource->usrp->daughterboard_id(0)) << ", ";
		result << "RX B: " << usrp_dbid_to_string(usrpSource->usrp->daughterboard_id(1)) << ", ";
		delete usrpSource;
	} else {
		result << "RX A: unknown, RX B: unknown, ";
	}
	rxDaughterboards = result.str();
	result.str("");

	if (usrpSink) {
		result << "TX A: " << usrp_dbid_to_string(usrpSink->usrp->daughterboard_id(0)) << ", ";
		result << "TX B: " << usrp_dbid_to_string(usrpSink->usrp->daughterboard_id(1)) << ".";
		delete usrpSink;
	} else {
		result << "TX A: unknown, TX B: unknown, ";
	}
	txDaughterboards = result.str();

	return;
}


/**
 * Local helper function to extract information about subdevices from a db_base_sptr into our
 * own subdevice structures
 */
void
extractSubdeviceInfo(
	const vector <vector <db_base_sptr>> &subdevs,
	subdevice &sideASubdev1, subdevice &sideASubdev2,
	subdevice &sideBSubdev1, subdevice &sideBSubdev2
) {
	// Extract data from the db_base_sptr objects: Side A
	if (subdevs[0].size() >= 1) {
		fetchDataFromUsrpDb(subdevs[0][0], sideASubdev1);
	} else {
		sideASubdev1.name = "Not available.";
	}
	if (subdevs[0].size() >= 2) {
		fetchDataFromUsrpDb(subdevs[0][1], sideASubdev2);
	} else {
		sideASubdev2.name = "Not available.";
	}

	// Extract data from the db_base_sptr objects: Side A
	if (subdevs[1].size() >= 1) {
		fetchDataFromUsrpDb(subdevs[1][0], sideBSubdev1);
	} else {
		sideBSubdev1.name = "Not available.";
	}
	if (subdevs[1].size() >= 2) {
		fetchDataFromUsrpDb(subdevs[1][1], sideBSubdev2);
	} else {
		sideBSubdev2.name = "Not available.";
	}
}


/**
 * Creates strings with the names of all subdevices on the Nth USRP Board on this computer.
 * Also, creates strings with the gain ranges of the subdevices.
 */
void
UsrpHelper::subdeviceInfo(
	int boardNumber, bool isTx,
	subdevice &sideASubdev1, subdevice &sideASubdev2,
	subdevice &sideBSubdev1, subdevice &sideBSubdev2
) {
	ostringstream result;

	// Clear subdevices, in case something else was written into them before
	sideASubdev1 = subdevice();
	sideASubdev2 = subdevice();
	sideBSubdev1 = subdevice();
	sideBSubdev2 = subdevice();

	// Validate range of the board number
	if (boardNumber < 0 || boardNumber > 1023) {
		result << "Unknown (Board Number #" << boardNumber << " out of range).";
		sideASubdev1.name = result.str();
		sideASubdev2.name = result.str();
		sideBSubdev1.name = result.str();
		sideBSubdev2.name = result.str();
		return;
	}

	// Create an UsrpSimulinkSource object
	try {
		if (isTx) {
			UsrpSimulinkSink *usrp = new UsrpSimulinkSink(boardNumber);
			extractSubdeviceInfo(
				usrp->usrp->db(),
				sideASubdev1, sideASubdev2,
				sideBSubdev1, sideBSubdev2
			);
			delete usrp;
		} else {
			UsrpSimulinkSource *usrp = new UsrpSimulinkSource(boardNumber);
			extractSubdeviceInfo(
				usrp->usrp->db(),
				sideASubdev1, sideASubdev2,
				sideBSubdev1, sideBSubdev2
			);
			delete usrp;
		}
	} catch (std::runtime_error E) {
		result << "Unknown (USRP Board #" << boardNumber << " not found).";
		sideASubdev1.name = result.str();
		sideASubdev2.name = result.str();
		sideBSubdev1.name = result.str();
		sideBSubdev2.name = result.str();
		return;
	}
}


/**
 * Returns a string with the sampling rate and USB data rate for a given decimation rate and
 * sample width.
 */
string
UsrpHelper::samplingRateAndBandwidthInfo(bool isTx, int decimation, int channelCount, int sampleWidth)
{
	ostringstream result;
	double freq = 0;


	// Check channel count
	if (channelCount < 1 || channelCount > 4 || (isTx && channelCount > 2)) {
		return "Number of channels out of range.";
	}

	// Check sample width
	if (sampleWidth != 1 && sampleWidth != 2) {
		return "Sample width out of range.";
	}

	// Oscillator frequency (hardcoded)
	if (isTx) {
		freq = 128e6;
	} else {
		freq = 64e6;
	}

	// Generate the output
	result << channelCount << " Channels: ";
	result << channelCount * (freq / 1000 / decimation) << " kSps, ";
	result << channelCount * (freq / 1024 * 2 * sampleWidth / decimation ) << " kB/s, ";
	result << channelCount * (freq / 1024 / 1024 * 2 * sampleWidth * 8 / decimation )
		<< " Mbit/s."
	;

	return result.str();
}


/**
 * Converts "Side A" or "Side B" into 0 or 1
 */
int
UsrpHelper::sideString2Side(wstring sideString, bool displayWarning)
{
	int side = 0;

	if (sideString == L"Side A") {
		//side = 0;
	} else if (sideString == L"Side B") {
		side = 1;
	} else {
		if (displayWarning) {
			printf("Unknown 'side' specified. Using Side A.\n");
		} else {
			side = -1;
		}
	}

	return side;
}


/**
 * Converts "8 Bit" or "16 Bit" into 1 or 2
 */
int
UsrpHelper::sampleWidthString2SampleWidth(string sampleWidthString, bool displayWarning)
{
	int sampleWidth = 2;

	if (sampleWidthString == "16 Bit") {
		//sampleWidth = 2;
	} else if (sampleWidthString == "8 Bit") {
		sampleWidth = 1;
	} else {
		if (displayWarning) {
			printf("Unknown 'sample_width' specified. Using 16 Bit.\n");
		} else {
			sampleWidth = -1;
		}
	}

	return sampleWidth;
}


/**
 * Converts "usrp_sink" in 'true'
 */
bool
UsrpHelper::typeString2RxType(string typeString, bool displayWarning)
{
	bool isTx = false;

	if (typeString == "usrp_sink") {
		isTx = true;
	} else if (typeString != "usrp_source") {
		if (displayWarning) {
			printf("Unknown block type specified: '%s'. Assuming an RX block.\n", typeString);
		}
	}

	return isTx;
}
