/**
 * Helper file for fetching information from the USRP hardware
 *
 * @version:	$Id: UsrpHelper.h 220 2009-05-15 07:39:45Z nagel $
 * @author:		Enno Klasing <klasing@int.uni-karlsruhe.de>
 * @copyright:	Institut fr Nachrichtentechnik, http://www.int.uni-karlsruhe.de
 * @license:	GNU General Public License, Version 3
 */

#include <string>


namespace UsrpHelper
{
	/**
	 * Struct with information about one single subdevice on a USRP daughterboard
	 */
	struct subdevice {
		std::string name;
		std::string gainRange;
		std::string frequencyRange;
		subdevice();
	};

	/**
	 * Returns a string with the serial number of the Nth USRP Board on this computer.
	 *
	 * @param	int			Nth USRP board on this computer.
	 */
	std::string boardInfo(int boardNumber);

	/**
	 * Collects two strings with the names of all daughterboards on the Nth USRP Board on this
	 * computer.
	 *
	 * @param	int			Nth USRP board on this computer.
	 * @param	string		Names of the RX daughterboards.
	 * @param	string		Names of the TX daughterboards.
	 */
	void daughterboardInfo(
		int boardNumber, std::string &rxDaughterboards, std::string &txDaughterboards
	);

	/**
	 * Creates strings with the names of all subdevices on the Nth USRP Board on this computer.
	 * Also, creates strings with the gain ranges of the subdevices.
	 *
	 * @param	int			Nth USRP board on this computer.
	 * @param	bool		True if this function shall operate on a UsrpSimulinkSink object,
	 *						false for a UsrpSimulinkSource object.
	 * @param	4*subdeviceInfo
	 *						Struct which will be populated with the name etc. of the subdevice
	 */
	void subdeviceInfo(
		int boardNumber,
		bool isTx,
		subdevice &sideASubdev1,
		subdevice &sideASubdev2,
		subdevice &sideBSubdev1,
		subdevice &sideBSubdev2
	);

	/**
	 * Returns a string with the sampling rate and USB data rate for a given decimation rate and
	 * sample width.
	 *
	 * @param	bool		True if this function shall operate on a UsrpSimulinkSink object,
	 *						false for a UsrpSimulinkSource object.
	 * @param	int			Decimation factor.
	 * @param	int			Number of channels that are transferred from the USRP to the Host.
	 * @param	int			Sample width (Bytes/Sample).
	 */
	std::string samplingRateAndBandwidthInfo(
		bool isTx, int decimation, int channelCount, int sampleWidth
	);

	/**
	 * Converts "Side A" or "Side B" into 0 or 1
	 *
	 * @param	wstring		Side.
	 * @param	bool		True if a warning shall be printed into Matlab when no valid side is
	 *						given.
	 */
	int sideString2Side(std::wstring sideString, bool displayWarning = true);

	/**
	 * Converts "8 Bit" or "16 Bit" into 1 or 2
	 *
	 * @param	string		Sample width.
	 * @param	bool		True if a warning shall be printed into Matlab when no valid sample
	 *						width is given.
	 */
	int sampleWidthString2SampleWidth(std::string sampleWidthString, bool displayWarning = true);

	/**
	 * Converts "usrp_sink" in 'true'
	 *
	 * @param	string		Type of the block.
	 * @param	bool		True if a warning shall be printed into Matlab when no valid type
	 *						is given.
	 */
	bool typeString2RxType(std::string typeString, bool displayWarning = true);
};
