/**
 * Interface between Matlab and the USRP, the USRP is used as a data sink
 *
 * This class extends the abstract base class usrp_sink_base and provides functions
 * for writing data from Simulink signals into the USRP.
 *
 * @version:	$Id: UsrpSimulinkSink.cpp 220 2009-05-15 07:39:45Z nagel $
 * @author:		Enno Klasing <klasing@int.uni-karlsruhe.de>
 * @copyright:	Institut fr Nachrichtentechnik, http://www.int.uni-karlsruhe.de
 * @license:	GNU General Public License, Version 3
 */

#include "UsrpSimulinkSink.h"
#include <usrp_bytesex.h>


/**
 * Constructor
 */
UsrpSimulinkSink::UsrpSimulinkSink(
	int which_board,
	unsigned int interp_rate,
	int nchan,
	int mux,
	int fusb_block_size,
	int fusb_nblocks,
	const std::string fpga_filename,
	const std::string firmware_filename
) :
	buffer(NULL),
	bufferSize(0),
	channelCount(0)
{
	usrp = usrp_standard_tx::make(
		which_board, interp_rate, nchan, mux,
		fusb_block_size, fusb_nblocks,
		fpga_filename, firmware_filename
	);
	if (usrp == 0) {
		throw std::runtime_error ("Can't open usrp.");
	}

}


/**
 * Destructor
 */
UsrpSimulinkSink::~UsrpSimulinkSink()
{
	// Deallocate any memory allocated in transfer()
	if (buffer != NULL) {
		free(buffer);
		buffer = NULL;
		bufferSize = 0;
	}
}


/**
 * Allocates storage for the local buffer of this class.
 */
bool
UsrpSimulinkSink::allocateBuffer(unsigned int nSamples)
{
	// Check that this operation writes only multiples of 512 Bytes to the USRP
	channelCount = usrp->nchannels();
	unsigned int nBytes = nSamples * 2 * bytesPerSample * channelCount;
	if (nBytes % 512 != 0) {
		return false;
	}

	// If the size of the buffer for reading from the USRP has changed, deallocate and reallocate
	// the memory (instead of realloc, since we don't need the old data anyway and thus can save
	// one copy operation of the whole memory block).
	if (buffer == NULL) {
		buffer = malloc(nBytes);
		bufferSize = nBytes;
	} else if (bufferSize != nBytes) {
		free(buffer);
		buffer = malloc(nBytes);
		bufferSize = nBytes;
	}

	return true;
}


/**
 * Writes a stream of complex-valued samples from the local buffer of this class
 * to the USRP
 */
bool
UsrpSimulinkSink::transfer()
{
	if (buffer == NULL) {
		return false;
	}

	bool underrun = false;

	// Write samples to the USRP
	int bytesWritten = usrp->write(buffer, bufferSize, &underrun);

	// Check for Underruns
	if (underrun) {
		fputs ("uU", stderr);
	}

	// Board unplugged/powered down or some other error?
	if (bytesWritten != bufferSize) {
		return false;
	}

	return true;
}


/**
 * Copies data from the given Simulink signal into the local buffer
 */
template <typename signal_T>
bool
UsrpSimulinkSink::copySimulinkSignal(
	unsigned int nSamples, unsigned int channelNum, const void *data
) {
	// Check that the local buffer is of the correct size
	unsigned int nSamplesTotal = nSamples * channelCount;
	if (nSamplesTotal * bytesPerSample * 2 != bufferSize) {
		return false;
	}

	// Fetch a pointer to the signal
	const signal_T *signalData = static_cast <const signal_T *> (data);

	// Fetch a pointer to the samples
	short *bufferData = static_cast <short *> (buffer);

	// Copy samples
	unsigned int k = 0;
	for (unsigned int i = channelNum; i < nSamplesTotal; i += channelCount) {
		bufferData[2*i    ] = host_to_usrp_short(static_cast <short> (signalData[k  ].re));
		bufferData[2*i + 1] = host_to_usrp_short(static_cast <short> (signalData[k++].im));
	}

	return true;
}


/**
 * Copies data from the two given Simulink signals into the local buffer
 */
template <typename signal_T>
bool
UsrpSimulinkSink::copySimulinkSignal(
	unsigned int nSamples, unsigned int channelNum, const void *dataReal, const void *dataImag
) {
	// Check that the local buffer is of the correct size
	unsigned int nSamplesTotal = nSamples * channelCount;
	if (nSamplesTotal * bytesPerSample * 2 != bufferSize) {
		return false;
	}

	// Fetch a pointer to the signal
	const signal_T *signalDataReal = static_cast <const signal_T *> (dataReal);
	const signal_T *signalDataImag = static_cast <const signal_T *> (dataImag);

	// Fetch a pointer to the samples
	short *bufferData = static_cast <short *> (buffer);

	// Copy samples
	unsigned int k = 0;
	for (unsigned int i = channelNum; i < nSamplesTotal; i += channelCount) {
		bufferData[2*i    ] = host_to_usrp_short(static_cast <short> (signalDataReal[k  ]));
		bufferData[2*i + 1] = host_to_usrp_short(static_cast <short> (signalDataImag[k++]));
	}

	return true;
}


/**
 * Explicit instantiations of the templated functions, since they are used in a different
 * compilation unit, and therefore can not be instantiated automatically.
 */
template bool
UsrpSimulinkSink::copySimulinkSignal <CINT8_T> (
	unsigned int nSamples, unsigned int channelNum, const void *data
);
template bool
UsrpSimulinkSink::copySimulinkSignal <CINT16_T> (
	unsigned int nSamples, unsigned int channelNum, const void *data
);
template bool
UsrpSimulinkSink::copySimulinkSignal <CINT32_T> (
	unsigned int nSamples, unsigned int channelNum, const void *data
);
template bool
UsrpSimulinkSink::copySimulinkSignal <CREAL_T> (
	unsigned int nSamples, unsigned int channelNum, const void *data
);
template bool
UsrpSimulinkSink::copySimulinkSignal <CREAL32_T> (
	unsigned int nSamples, unsigned int channelNum, const void *data
);
template bool
UsrpSimulinkSink::copySimulinkSignal <INT8_T> (
	unsigned int nSamples, unsigned int channelNum, const void *dataReal, const void *dataImag
);
template bool
UsrpSimulinkSink::copySimulinkSignal <INT16_T> (
	unsigned int nSamples, unsigned int channelNum, const void *dataReal, const void* dataImag
);
template bool
UsrpSimulinkSink::copySimulinkSignal <INT32_T> (
	unsigned int nSamples, unsigned int channelNum, const void *dataReal, const void* dataImag
);
template bool
UsrpSimulinkSink::copySimulinkSignal <REAL_T> (
	unsigned int nSamples, unsigned int channelNum, const void *dataReal, const void* dataImag
);
template bool
UsrpSimulinkSink::copySimulinkSignal <REAL32_T> (
	unsigned int nSamples, unsigned int channelNum, const void *dataReal, const void* dataImag
);
