/**
 * Interface between Matlab and the USRP, the USRP is used as a data sink
 *
 * This class extends the abstract base class usrp_sink_base and provides functions
 * for writing data from Simulink signals into the USRP.
 *
 * @version:	$Id: UsrpSimulinkSink.h 220 2009-05-15 07:39:45Z nagel $
 * @author:		Enno Klasing <klasing@int.uni-karlsruhe.de>
 * @copyright:	Institut fr Nachrichtentechnik, http://www.int.uni-karlsruhe.de
 * @license:	GNU General Public License, Version 3
 */

#include <usrp_standard.h>
#include <simstruc.h>


class UsrpSimulinkSink
{
private:
	void *buffer;						// Pointer to the buffer of the raw bytes.
	unsigned int bufferSize;			// Length of the allocated buffer in Bytes
	static const int bytesPerSample = 2;// Bytes per sample to the USRP (sinks only support 2)
	int channelCount;					// Number of channels that will be written to the USRP.
										// This equals this->usrp->nchannels(), but saves us a
										// few function calls.

public:
	usrp_standard_tx_sptr usrp;			// The actual usrp object

	/**
	 * Constructor
	 */
	UsrpSimulinkSink(
		int which_board,
		unsigned int interp_rate = 512,
		int nchan = 1,
		int mux = -1,
		int fusb_block_size = 0,
		int fusb_nblocks = 0,
		const std::string fpga_filename = "",
		const std::string firmware_filename = ""
	);

	/**
	 * Destructor
	 */
	~UsrpSimulinkSink();

	/**
	 * Allocates storage for the local buffer of this class.
	 *
	 * @param	unsigned int	Number of complex samples to use in this operation. This must be
	 *							a multiple of 512 / sizeof("complex USRP sample"), which results in
	 *							multiples of 128 complex 16-Bit or 256 complex 8-Bit samples.
	 */
	bool allocateBuffer(unsigned int nSamples);

	/**
	 * Writes a stream of complex-valued samples from the local buffer of this class
	 * to the USRP
	 *
	 * @return	bool			True on success (even if an USRP overrun/underrun occured), false
	 *							otherwise.
	 */
	bool transfer();

	/**
	 * Copies data from the given Simulink signal into the local buffer
	 *
	 * @param	unsigned int	Number of samples.
	 * @param	unsigned int	Number of the channel that is to be written into the buffer.
	 * @param	void			Pointer to the begin of the signal vector.
	 */
	template <typename signal_T>
	bool copySimulinkSignal(
		unsigned int nSamples, unsigned int channelNum, const void *data
	);

	/**
	 * Copies data from the two given Simulink signals into the local buffer
	 *
	 * @param	unsigned int	Number of samples.
	 * @param	unsigned int	Number of the channel that is to be written into the buffer.
	 * @param	void			Pointer to the begin of the signal vector, real part.
	 * @param	void			Pointer to the begin of the signal vector, imaginary
	 *							part.
	 */
	template <typename signal_T>
	bool copySimulinkSignal(
		unsigned int nSamples, unsigned int channelNum, const void *dataReal, const void *dataImag
	);

	/**
	 * Typedefs
	 */
	typedef	bool (UsrpSimulinkSink::* copyFunctionPointerComplex) (
		unsigned int nSamples, unsigned int channelNum, const void *data
	);
	typedef	bool (UsrpSimulinkSink::* copyFunctionPointerReal) (
		unsigned int nSamples, unsigned int channelNum, const void *dataReal, const void *dataImag
	);
};
