/**
 * Interface between Matlab and the USRP, the USRP is used as a data source
 *
 * This class extends the abstract base class usrp_source_base and provides functions
 * for reading data from the USRP into Simulink signals.
 *
 * @version:	$Id: UsrpSimulinkSource.cpp 220 2009-05-15 07:39:45Z nagel $
 * @author:		Enno Klasing <klasing@int.uni-karlsruhe.de>
 * @copyright:	Institut fr Nachrichtentechnik, http://www.int.uni-karlsruhe.de
 * @license:	GNU General Public License, Version 3
 */

#include "UsrpSimulinkSource.h"
#include <usrp_bytesex.h>


/**
 * Constructor
 */
UsrpSimulinkSource::UsrpSimulinkSource(
	int which_board,
	unsigned int decim_rate,
	int nchan,
	int mux,
	int mode,
	int fusb_block_size,
	int fusb_nblocks,
	const std::string fpga_filename,
	const std::string firmware_filename
) :
	buffer(NULL),
	bufferSize(0),
	bytesPerSample(0),
	channelCount(0)
{
	usrp = usrp_standard_rx::make(
		which_board, decim_rate, nchan, mux, mode,
		fusb_block_size, fusb_nblocks,
		fpga_filename, firmware_filename
	);
	if (usrp == 0) {
		throw std::runtime_error ("Can't open usrp.");
	}
}


/**
 * Destructor
 */
UsrpSimulinkSource::~UsrpSimulinkSource()
{
	// Deallocate any memory allocated in transfer()
	if (buffer != NULL) {
		free(buffer);
		buffer = NULL;
		bufferSize = 0;
	}
}


/**
 * Allocates storage for the local buffer of this class.
 */
bool
UsrpSimulinkSource::allocateBuffer(unsigned int nSamples)
{
	// Check that this operation fetches only multiples of 512 Bytes from the USRP
	bytesPerSample = usrp->format_width(usrp->format()) / 8;
	channelCount = usrp->nchannels();
	unsigned int nBytes = nSamples * 2 * bytesPerSample * channelCount;
	if (nBytes % 512 != 0) {
		return false;
	}

	// If the size of the buffer for reading from the USRP has changed, deallocate and reallocate
	// the memory (instead of realloc, since we don't need the old data anyway and thus can save
	// one copy operation of the whole memory block).
	if (buffer == NULL) {
		buffer = malloc(nBytes);
		bufferSize = nBytes;
	} else if (bufferSize != nBytes) {
		free(buffer);
		buffer = malloc(nBytes);
		bufferSize = nBytes;
	}

	return true;
}


/**
 * Reads a stream of complex-valued samples from the USRP and writes it into the
 * supplied Simulink output port signal
 */
bool
UsrpSimulinkSource::transfer()
{
	if (buffer == NULL) {
		return false;
	}

	bool overrun = false;

	// Read samples from the USRP
	int bytesRead = usrp->read(buffer, bufferSize, &overrun);

	// Check for Overruns
	if (overrun) {
		fputs ("uO", stderr);
	}

	// Fewer Bytes returned than requested?
	if (bytesRead != bufferSize) {
		fprintf (stderr, "usrp_source: short read.  Expected %d, got %d\n", bufferSize, bytesRead);
		return false;
	}

	return true;
}


/**
 * Copies data from the local buffer into the given Simulink signal
 */
template <typename signal_T>
bool
UsrpSimulinkSource::copySimulinkSignal(
	unsigned int nSamples, unsigned int channelNum, void *data
) {
	// Check that the local buffer is of the correct size
	unsigned int nSamplesTotal = nSamples * channelCount;
	if (nSamplesTotal * bytesPerSample * 2 != bufferSize) {
		return false;
	}

	// Fetch a pointer to the signal
	signal_T *signalData = static_cast <signal_T *> (data);

	// Copy the data, according to the USRP data format
	unsigned int k = 0;
	if (bytesPerSample == 1) {
		// Fetch a pointer to the samples
		const signed char *bufferData = static_cast <const signed char *> (buffer);

		// Copy samples
		for (unsigned int i = channelNum; i < nSamplesTotal; i += channelCount) {
			signalData[k  ].re = static_cast <short> (bufferData[2*i + 0] << 8);
			signalData[k++].im = static_cast <short> (bufferData[2*i + 1] << 8);
		}
	} else if (bytesPerSample == 2) {
		// Fetch a pointer to the samples
		const short *bufferData = static_cast <const short *> (buffer);

		// Copy samples
		for (unsigned int i = channelNum; i < nSamplesTotal; i += channelCount) {
			signalData[k  ].re = usrp_to_host_short(bufferData[2*i + 0]);
			signalData[k++].im = usrp_to_host_short(bufferData[2*i + 1]);
		}
	} else {
		// Unknown USRP data format!
		return false;
	}

	return true;
}


/**
 * Copies data from the local buffer into the given Simulink signal
 * Explicit template specialization for CINT8_T
 */
template <>
bool
UsrpSimulinkSource::copySimulinkSignal <CINT8_T> (
	unsigned int nSamples, unsigned int channelNum, void *data
) {
	// Check that the local buffer is of the correct size
	unsigned int nSamplesTotal = nSamples * channelCount;
	if (nSamplesTotal * bytesPerSample * 2 != bufferSize) {
		return false;
	}

	// Fetch a pointer to the signal
	CINT8_T *signalData = static_cast <CINT8_T *> (data);

	// Copy the data, according to the USRP data format
	unsigned int k = 0;
	if (bytesPerSample == 1) {
		// Fetch a pointer to the samples
		const signed char *bufferData = static_cast <const signed char *> (buffer);

		// Copy samples
		for (unsigned int i = channelNum; i < nSamplesTotal; i += channelCount) {
			signalData[k  ].re = bufferData[2*i + 0];
			signalData[k++].im = bufferData[2*i + 1];
		}
	} else if (bytesPerSample == 2) {
		// Fetch a pointer to the samples
		const short *bufferData = static_cast <const short *> (buffer);

		// Copy samples
		for (unsigned int i = channelNum; i < nSamplesTotal; i += channelCount) {
			signalData[k  ].re = static_cast <INT8_T> (usrp_to_host_short(bufferData[2*i + 0]));
			signalData[k++].im = static_cast <INT8_T> (usrp_to_host_short(bufferData[2*i + 1]));
		}
	} else {
		// Unknown USRP data format!
		return false;
	}

	return true;
}


/**
 * Copies data from the local buffer into the two given Simulink signals
 */
template <typename signal_T>
bool
UsrpSimulinkSource::copySimulinkSignal(
	unsigned int nSamples, unsigned int channelNum, void *dataReal, void *dataImag
) {
	// Check that the local buffer is of the correct size
	unsigned int nSamplesTotal = nSamples * channelCount;
	if (nSamplesTotal * bytesPerSample * 2 != bufferSize) {
		return false;
	}

	// Fetch a pointer to the signal
	signal_T *signalDataReal = static_cast <signal_T *> (dataReal);
	signal_T *signalDataImag = static_cast <signal_T *> (dataImag);

	// Copy the data, according to the USRP data format
	unsigned int k = 0;
	if (bytesPerSample == 1) {
		// Fetch a pointer to the samples
		const signed char *bufferData = static_cast <const signed char *> (buffer);

		// Copy samples
		for (unsigned int i = channelNum; i < nSamplesTotal; i += channelCount) {
			signalDataReal[k  ] = static_cast <short> (bufferData[2*i + 0] << 8);
			signalDataImag[k++] = static_cast <short> (bufferData[2*i + 1] << 8);
		}
	} else if (bytesPerSample == 2) {
		// Fetch a pointer to the samples
		const short *bufferData = static_cast <const short *> (buffer);

		// Copy samples
		for (unsigned int i = channelNum; i < nSamplesTotal; i += channelCount) {
			signalDataReal[k  ] = usrp_to_host_short(bufferData[2*i + 0]);
			signalDataImag[k++] = usrp_to_host_short(bufferData[2*i + 1]);
		}
	} else {
		// Unknown USRP data format!
		return false;
	}

	return true;
}


/**
 * Copies data from the local buffer into the two given Simulink signals
 * Explicit template specialization for CINT8_T
 */
template <>
bool
UsrpSimulinkSource::copySimulinkSignal <INT8_T> (
	unsigned int nSamples, unsigned int channelNum, void *dataReal, void *dataImag
) {
	// Check that the local buffer is of the correct size
	unsigned int nSamplesTotal = nSamples * channelCount;
	if (nSamplesTotal * bytesPerSample * 2 != bufferSize) {
		return false;
	}

	// Fetch a pointer to the signal
	INT8_T *signalDataReal = static_cast <INT8_T *> (dataReal);
	INT8_T *signalDataImag = static_cast <INT8_T *> (dataImag);

	// Copy the data, according to the USRP data format
	unsigned int k = 0;
	if (bytesPerSample == 1) {
		// Fetch a pointer to the samples
		const signed char *bufferData = static_cast <const signed char *> (buffer);

		// Copy samples
		for (unsigned int i = channelNum; i < nSamplesTotal; i += channelCount) {
			signalDataReal[k  ] = bufferData[2*i + 0];
			signalDataImag[k++] = bufferData[2*i + 1];
		}
	} else if (bytesPerSample == 2) {
		// Fetch a pointer to the samples
		const short *bufferData = static_cast <const short *> (buffer);

		// Copy samples
		for (unsigned int i = channelNum; i < nSamplesTotal; i += channelCount) {
			signalDataReal[k  ] = static_cast <INT8_T> (usrp_to_host_short(bufferData[2*i + 0]));
			signalDataImag[k++] = static_cast <INT8_T> (usrp_to_host_short(bufferData[2*i + 1]));
		}
	} else {
		// Unknown USRP data format!
		return false;
	}

	return true;
}


/**
 * Explicit instantiations of the templated functions, since they are used in a different
 * compilation unit, and therefore can not automatically be instantiated.
 */
template bool
UsrpSimulinkSource::copySimulinkSignal <CINT16_T> (
	unsigned int nSamples, unsigned int channelNum, void *data
);
template bool
UsrpSimulinkSource::copySimulinkSignal <CINT32_T> (
	unsigned int nSamples, unsigned int channelNum, void *data
);
template bool
UsrpSimulinkSource::copySimulinkSignal <CREAL_T> (
	unsigned int nSamples, unsigned int channelNum, void *data
);
template bool
UsrpSimulinkSource::copySimulinkSignal <CREAL32_T> (
	unsigned int nSamples, unsigned int channelNum, void *data
);
template bool
UsrpSimulinkSource::copySimulinkSignal <INT16_T> (
	unsigned int nSamples, unsigned int channelNum, void *dataReal, void* dataImag
);
template bool
UsrpSimulinkSource::copySimulinkSignal <INT32_T> (
	unsigned int nSamples, unsigned int channelNum, void *dataReal, void* dataImag
);
template bool
UsrpSimulinkSource::copySimulinkSignal <REAL_T> (
	unsigned int nSamples, unsigned int channelNum, void *dataReal, void* dataImag
);
template bool
UsrpSimulinkSource::copySimulinkSignal <REAL32_T> (
	unsigned int nSamples, unsigned int channelNum, void *dataReal, void* dataImag
);
