/**
 * Interface between Matlab and the USRP, the USRP is used as a data source
 *
 * This class extends the abstract base class usrp_source_base and provides functions
 * for reading data from the USRP into Simulink signals.
 *
 * @version:	$Id: UsrpSimulinkSource.h 220 2009-05-15 07:39:45Z nagel $
 * @author:		Enno Klasing <klasing@int.uni-karlsruhe.de>
 * @copyright:	Institut fr Nachrichtentechnik, http://www.int.uni-karlsruhe.de
 * @license:	GNU General Public License, Version 3
 */

#include <usrp_standard.h>
#include <simstruc.h>


class UsrpSimulinkSource
{
private:
	void *buffer;						// Pointer to the buffer of the raw bytes.
	unsigned int bufferSize;			// Length of the allocated buffer in Bytes
	int bytesPerSample;					// Bytes per sample from the USRP (1..2)
	int channelCount;					// Number of channels that were read from the USRP.
										// This equals this->nchannels(), but saves us a few
										// function calls.

public:
	usrp_standard_rx_sptr usrp;			// The actual usrp object

	/**
	 * Constructor
	 */
	UsrpSimulinkSource(
		int which_board,
		unsigned int decim_rate = 256,
		int nchan = 1,
		int mux = -1,
		int mode = 0,
		int fusb_block_size = 0,
		int fusb_nblocks = 0,
		const std::string fpga_filename = "",
		const std::string firmware_filename = ""
	);

	/**
	 * Destructor
	 */
	~UsrpSimulinkSource();

	/**
	 * Allocates storage for the local buffer of this class.
	 *
	 * @param	unsigned int	Number of complex samples to use in this operation. This must be
	 *							a multiple of 512 / sizeof("complex USRP sample"), which results in
	 *							multiples of 128 complex 16-Bit or 256 complex 8-Bit samples.
	 */
	bool allocateBuffer(unsigned int nSamples);

	/**
	 * Reads a stream of complex-valued samples from the USRP and writes it into the
	 * local buffer of this class
	 *
	 * @return	bool			True on success (even if an USRP overrun/underrun occured), false
	 *							otherwise.
	 */
	bool transfer();

	/**
	 * Copies data from the local buffer into the given Simulink signal
	 *
	 * @param	unsigned int	Number of samples.
	 * @param	unsigned int	Number of the channel that is to be extracted from the buffer.
	 * @param	void			Pointer to the begin of the signal vector.
	 */
	template <typename signal_T>
	bool copySimulinkSignal(
		unsigned int nSamples, unsigned int channelNum, void *data
	);

	/**
	 * Copies data from the local buffer into the two given Simulink signals
	 *
	 * @param	unsigned int	Number of samples.
	 * @param	unsigned int	Number of the channel that is to be extracted from the buffer.
	 * @param	void			Pointer to the begin of the signal vector, real part.
	 * @param	void			Pointer to the begin of the signal vector, imaginary
	 *							part.
	 */
	template <typename signal_T>
	bool copySimulinkSignal(
		unsigned int nSamples, unsigned int channelNum, void *dataReal, void *dataImag
	);

	/**
	 * Typedefs
	 */
	typedef	bool (UsrpSimulinkSource::* copyFunctionPointerComplex) (
		unsigned int nSamples, unsigned int channelNum, void *data
	);
	typedef	bool (UsrpSimulinkSource::* copyFunctionPointerReal) (
		unsigned int nSamples, unsigned int channelNum, void *dataReal, void *dataImag
	);
};

/**
 * Explicit template specialization declaration for [C]INT8_T. If this occurs inside of the class
 * definition, Visual Studio does not recognize UsrpSimulinkSource as a class.
 */
template <>
bool
UsrpSimulinkSource::copySimulinkSignal <CINT8_T> (
	unsigned int nSamples, unsigned int channelNum, void *data
);
template <>
bool
UsrpSimulinkSource::copySimulinkSignal <INT8_T> (
	unsigned int nSamples, unsigned int channelNum, void *dataReal, void *dataImag
);
