/**
 * Helper file to create a named pipe, which will be forwarded to mexPrintf.
 *
 * This is done because otherwise the output of fputs() or fprintf(stderr/stdout)
 * won't be shown in Matlab.
 *
 * @version:	$Id: log2namedPipe.cpp 220 2009-05-15 07:39:45Z nagel $
 * @author:		Enno Klasing <klasing@int.uni-karlsruhe.de>
 * @copyright:	Institut fr Nachrichtentechnik, http://www.int.uni-karlsruhe.de
 * @license:	GNU General Public License, Version 3
 */

#include "log2namedPipe.h"
#include <mex.h>
#include <string>
#include <windows.h>
#include <errno.h>

using namespace std;


#define MEX_ERROR_MESSAGE(message, ...) mexPrintf(##message " In function %s, line %u.\n", __VA_ARGS__, __FUNCTION__, __LINE__)


/**
 * Constants
 */
const int NAMED_PIPE_NAME_MAX_LENGTH = 256;
enum namedPipeType {
	STDOUT = 0,
	STDERR,
	// NAMED_PIPE_TYPE_LENGTH must be the last in this enum to correctly identify the length of the threadHandles array
	NAMED_PIPE_TYPE_LENGTH
};

/**
 * Structs and Typedefs
 */
struct pipeListenerParameters {
	HANDLE			pipe;
	namedPipeType	type;
	string			name;
	bool			verbose;
};

/**
 * Global Variables
 */
// Holds handles to the created threads
static HANDLE threadHandles[NAMED_PIPE_TYPE_LENGTH];

// Handle to a MUTEX object so that the thread handles of the pipes can be modified safely.
static HANDLE mutex = NULL;

/**
 * Listens on a named pipe and prints the data which comes in through the pipe into Matlab.
 *
 * This function should be started as a new thread. Thus, the rest of the program can keep
 * running, while this functions blocks until the last pipe of this name has been closed.
 *
 * @param	LPVOID	Pointer to a pipeListenerParameters structure (only one parameter can be passed to a new thread). This thread deletes the memory of the structure after using it.
 *			HANDLE	pipe: Handle to the named pipe this thread should listen to.
 *			bool	copyToDebugger: True if all data coming through the pipe shall also be copied into a debugger.
 * @return	DWORD	Void, since this runs as a thread, but the thread's return value will not be captured.
 */
DWORD WINAPI
listenOnNamedPipe(LPVOID params)
{
	BOOL result = FALSE;
	DWORD bytesRead = 0;
	DWORD error = 0;
	HANDLE pipe = NULL;
	namedPipeType type = STDOUT;
	string pipeName;
	bool verbose = false;
	char buffer[1];
	buffer[0] = ' ';

	// Convert the parameters
	pipe = static_cast <pipeListenerParameters *> (params)->pipe;
	type = static_cast <pipeListenerParameters *> (params)->type;
	pipeName = static_cast <pipeListenerParameters *> (params)->name;
	verbose = static_cast <pipeListenerParameters *> (params)->verbose;
	delete params; // Allocated in createNamedPipeAndListenerThread(), deallocated here because it got passed into this function

	// Wait for a client to connect
	result = ConnectNamedPipe(pipe, NULL);
	if (!result && (error = GetLastError()) != ERROR_PIPE_CONNECTED) {
		MEX_ERROR_MESSAGE("ConnectNamedPipe on pipe %s (%u) failed. Last error code: %u.", pipeName.data(), pipe, error);
	} else {
		if (verbose) {
			mexPrintf("Received connection on named pipe %s (%u).\n", pipeName.data(), pipe);
		}

		// Read from the pipe as long as there is no error
		result = ReadFile(pipe, buffer, 1, &bytesRead, NULL);
		while (result) {
			mexPrintf("%c", buffer[0]);
			result = ReadFile(pipe, buffer, 1, &bytesRead, NULL);
		}

		// Check if there was an error different to "Broken pipe"
		error = GetLastError();
		if (error != ERROR_BROKEN_PIPE) {
			MEX_ERROR_MESSAGE("Reading from pipe %s (%u) failed. Last error code: %u.", pipeName.data(), pipe, error);
		}
	}

	// We're done, close the pipe
	if (!CloseHandle(pipe)) {
		error = GetLastError();
		MEX_ERROR_MESSAGE("CloseHandle on pipe %s (%u) failed. Last error code: %u.", pipeName.data(), pipe, error);
	} else {
		if (verbose) {
			mexPrintf("Closed named pipe %s (%u).\n", pipeName.data(), pipe);
		}
	}

	return false;
}


/**
 * Creates a named pipe and a new thread which listens on the server end of the pipe.
 * Redirects the specified output stream into this pipe.
 *
 * The data which comes in through the pipe will be printed into Matlab, using mexPrintf().
 *
 * @param	int			Type of the pipe: STDERR/STDOUT
 * @param	bool		True to print out status messages.
 * @return	bool		True on success
 */
bool
createNamedPipeAndListenerThread(const namedPipeType type, bool verbose)
{
	DWORD result = 0;
	HANDLE thread = NULL;
	HANDLE pipe = NULL;
	char pipeName[NAMED_PIPE_NAME_MAX_LENGTH];
	pipeListenerParameters *params = 0;

	// Check if there is already one thread for this output redirection running
	if (threadHandles[type] != NULL) {
		// If the thread has died, close it's handle
		if (WaitForSingleObject(threadHandles[type], 0) == WAIT_OBJECT_0) {
			CloseHandle(threadHandles[type]);
			threadHandles[type] = NULL;
		} else {
			return true;
		}
	}

	// Build the name of the named pipe
	sprintf_s(pipeName, "\\\\.\\pipe\\matlab_%u_%u", type, GetCurrentProcessId());

	// Create the named pipe
	pipe = CreateNamedPipeA(
		pipeName,
		PIPE_ACCESS_INBOUND,
		PIPE_TYPE_BYTE || PIPE_WAIT,
		1,
		0,
		1024,
		NMPWAIT_USE_DEFAULT_WAIT,
		NULL
	);

	// Pipe created successfully?
	if (pipe == INVALID_HANDLE_VALUE) {
		MEX_ERROR_MESSAGE("CreateNamedPipe() failed. pipeName = %s.", pipeName);
		return false;
	}

	// Create listener thread on this named pipe
	params = new pipeListenerParameters; // Gets deleted by the new thread (or here if the thread can not be created).
	params->pipe = pipe;
	params->type = type;
	params->name = pipeName;
	params->verbose = verbose;
	thread = CreateThread(
		NULL,
		0,
		listenOnNamedPipe,
		params,
		0,
		NULL
	);

	// Thread started successfully?
	if (thread == NULL) {
		MEX_ERROR_MESSAGE("CreateThread() failed.");
		delete params;
		return false;
	} else {
		// Save the thread's handle
		threadHandles[type] = thread;

		// Redirect the output stream into the pipe
		FILE *redirectedStream = 0;
		errno_t result;
		result = freopen_s(&redirectedStream, pipeName, "wb", (type == STDOUT ? stdout : stderr));
		if (result != 0 || !redirectedStream) {
			MEX_ERROR_MESSAGE(
				"Redirecting %s into named pipe %s failed.",
				(type == STDOUT ? "stdout" : "stderr"), pipeName
			);
			return false;
		}

		// Set the buffer mode to "No Buffering" for the selected stream
		if(setvbuf(redirectedStream, NULL, _IONBF, 0) != 0) {
			MEX_ERROR_MESSAGE("Deactivating buffer for named pipe %s failed.", pipeName);
		}
	}

	return true;
}

/**
 * Redirects stdout and stderr into two named pipes, which will call mexPrintf() for all data coming
 * it's way.
 */
bool
redirectStdoutAndStderr(bool registerMexAtExit, bool verbose)
{
	bool result = true;

	// Register our own mexAtExit function to close all redirected streams and to kill all created threads before unloading this library
	if (registerMexAtExit) {
		mexAtExit(deleteStdoutAndStderrRedirection);
	}

	// Creates an anonymous MUTEX object once, and initializes the array of thread handles to zero
	if (mutex == NULL) {
		mutex = CreateMutex(NULL, FALSE, NULL);
		for (int i = 0; i < NAMED_PIPE_TYPE_LENGTH; i++) {
			threadHandles[i] = NULL;
		}
	}

	// Lock the MUTEX
	if (WaitForSingleObject(mutex, INFINITE) != WAIT_OBJECT_0) {
		MEX_ERROR_MESSAGE("WaitForSingleObject (MUTEX) failed.");
	}

	// Create pipe for redirecting stdout
	if (!createNamedPipeAndListenerThread(STDOUT, verbose)) {
		result = false;
	}

	// Create pipe for redirecting stderr
	if (!createNamedPipeAndListenerThread(STDERR, verbose)) {
		result = false;
	}

	// Unlock the MUTEX
	if (!ReleaseMutex(mutex)) {
		MEX_ERROR_MESSAGE("ReleaseMutex() failed.");
	}

	return result;
}

/**
 * Closes stdout and stderr, thus effectively breaking the pipes and ending the pipe listener threads.
 */
void
deleteStdoutAndStderrRedirection()
{
	DWORD result = 0;

	// Check if redirectStdoutAndStderr() has been called at least once
	if (mutex == NULL) {
		return;
	}

	// Close the streams, so that the ReadFile operations in the listenOnNamedPipe threads will return an error
	fclose(stdout);
	fclose(stderr);

	// Lock the MUTEX
	result = WaitForSingleObject(mutex, INFINITE);
	if (result != WAIT_OBJECT_0) {
		MEX_ERROR_MESSAGE("WaitForSingleObject (MUTEX) failed.");
	}

	// Wait for the threads to exit. Otherwise, this DLL may get unloaded before the threads have terminated!
	for (int i = 0; i < NAMED_PIPE_TYPE_LENGTH; i++) {
		if (threadHandles[i]) {
			result = WaitForSingleObject(threadHandles[i], INFINITE);
			if (result != WAIT_OBJECT_0) {
				MEX_ERROR_MESSAGE("WaitForSingleObject (thread) failed.");
			} else {
				CloseHandle(threadHandles[i]);
			}
			threadHandles[i] = NULL;
		}
	}

	// Unlock the MUTEX
	if (!ReleaseMutex(mutex)) {
		MEX_ERROR_MESSAGE("ReleaseMutex() failed.");
	}
}
